#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals 		= 3
#pragma IgorVersion		= 8.0
#pragma ModuleName		= SpectraImageProfile
#pragma version			= 4.45
#pragma DefaultTab		= {3,20,4}		// set default tab width in Igor Pro 9 and later

#include <PopupWaveSelector>

static Constant kVersion = 4.45
static StrConstant kVersionDate = "12/2024"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21549																// the project node on IgorExchange
static StrConstant ksShortTitle = "Profile Generator"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//
//	SpectraProfile is a environment to create 1D profiles from 2D or 3D data sets in horizontal or
//	vertical direction. Various modifications to the 2D/3D data itself can be applied. Some of the
//	functionality relies on properties of the data used (such as scaling and column labels). Some
//	modification of functionality by the user is possible.
//	
//	keyboard shortcuts:
//	• s key			saves the current profile
//	• p key			creates and shows a snapshot
//	• u key			undo changes
//	• del key		deletes the current ROI section
//	• g key			toggle ‘Snap to Grid’
//	• m key			toggle live ‘Mouse Profile’ mode
//	• n key			toggle ‘Normalize to one’
//	• h key			switch to horizontal mode
//	• v key			switch to vertical mode
//	• z key			show / hide z profile
//	• ← → arrow keys in horizontal mode					change width of integration area (horizontal mode)
//	• ↑↓ arrow keys in vertical mode					change width of integration area (vertical mode)
//	• ↑↓ arrow keys in horizontal mode					move integration area up / down (horizontal mode)
//	• ← → arrow keys in vertical mode					move integration area left / right (vertical mode)
//	• page up, page down keys							switch to ‘Full Width’/ ‘Minimal Width’
//	• ctrl + arrow keys									changes are slower by 0.5 x the step size
//	• shift + arrow keys								changes are faster by 4 x the step size
//	• both ctrl & shift + arrow keys					changes exactly by one delta unit (data point)
//	• hold shift while saving a profile					omit region info from name (just add ‘_prH’ / ‘_prV’)
//	• hold alt while saving a profile					save the output to the source folder instead of the current folder
//	• hold shift while pressing ‘Align Features’		align to the top / right, instead of the bottom / left
//	• set cursors before pressing ‘Align Features’ 		aligns to the sub-region between the cursors
//	• hold ctrl while pressing a normalize button		hard normalization without any smoothing
//	• hold ctrl while using ‘< / > Shift’				tries to cut edges with partial data after conversion
//	• hold ctrl/alt while using ‘subtract profile’		subtract the profile while moving with increasing / decreasing y-axis values.
//														Example: Subtract BE-features in KE-spectra. Works only in horizontal mode.
//	• shift + one of the quick-profile menu entries		a prompt appears to enter a sub-range for integration
//	• shift + one of the quick-split menu entries		a prompt appears to enter a format string for wave naming
//
//
// 2021-02-16 - ver. 3.60:	Initial public release.
// 2021-07-22 - ver. 3.70:	Slightly adjusted the names in the main menu entries.
// 							Fixed wrong wave labels (row vs. col) for the split tool.
//							Added scaled positions to the wave name for the split tool.
//							Added keyboard shortcuts for Normalize Profile and Snap to Grid.
//							Added Accept & Quit button.
//							Added optional logging of data edits in the history.
//							Added control to set the smoothing factor for the align feature.
//							Moved all controls into an external panel.
//							Now the live profile under the mouse cursor gets normalized as well.
//							Fixed the way-too-big smoothing value when using the align feature.
//							Data units get displayed in the graph now.
// 2021-07-24 - ver. 3.71:	Improvements of matrix calculations thanks to suggestions by A.G. (WaveMetrics).
//							Added image contrast controls.
// 2021-10-26 - ver. 3.72:	When combining waves the source waves are recorded in the combined wave's notes now.
//							Added sub-step shifting to the Align feature.
// 2021-12-23 - ver. 3.80:	Made sure that sub-step alignment does not throw errors if the fit fails.
//							Fixed wrong panel size on windows with big fonts.
//							Align features with cursors present was not working correctly in some cases.
//							Added Align Edges feature which aligns features in the derivative.
//							Image profile now works with 3D data as well, either on the current layer or the merged image.
// 2022-03-23 - ver. 3.90:	Fixed error message when QuickProfile and QuickSplit are run with a folder selected.
//							Reduced ROI expansion via keyboard to single steps (unless snap-to-grid is activated).
//							Steps for repositioning and width changes via keyboard are now adapted to the data size.
//							Initial center value will be aligned to the grid for waves with less than 100 rows / columns, and honors zero.
//							Added Delete button to delete individual layers from 3D data.
//							Integration ranges are now indicated in the output name with 'to' instead of a hyphen/minus to avoid confusion.
//							Now multiple instances of the tool can be opened at the same time.
//							Added support for keyboard-shortcut (and other future) settings via external text file.
//							Axis split between image and profile can now be dragged to a new position.
// 2022-10-28 - ver. 4.00:	Added support for data units.
//							Fixed errors when selecting data folders instead of waves or when nothing is selected.
//							Now 3D volumes can be rotated around the three principal axes.
//							Axes are now reversed according to the sign of the delta-scaling in the respective direction.
//							Many tweaks and bug-fixes added.
//							---- Many thanks to Oleg Dyachok from Uppsala University for suggesting and testing the changes below ----
//							Holding both shift + control / command while using the arrow keys will change the center / width by exactly one unit.
//							Combining and stacking waves writes the input wave names as dimension labels.
//							Added display of multiple profiles generated from all layers of a 3D volume.
//							Added support for displaying dimension labels in the Image Profile tool.
//							Added color scale legend to the image.
// 2023-01-07 - ver. 4.05:	Added right-click menu to the Data Browser.
//							Added data-binning functionality.
//							Multi-profiles are now saved as well.
//							Dimension labels are rotated together with the data.
//							Added a 'Save as' button to save the input data under a new name (after edits etc.).
// 2023-03-19 - ver. 4.10:	Notes are now written into the output when splitting waves into rows / columns / layers.
//							Sub-range for integration can now be specified for Quick-Profile function.
//							Also, holding shift while invoking the menu summons a user interface to select the integration range.
//							Changed the meaning of the 'dim' parameter for the Quick Profile and Quick Split functions to mean the dimension to be integrated / split.
//							Fixed bug: The buttons of the inset panel could appear on different windows in some circumstances.
//							Fixed bug: Closing one instance broke updates in other instances.
//							Fixed bug: Some buttons were not sized properly for Mac.
// 2023-07-25 - ver. 4.20:	QProfile() and QSplit() now return the name of the created wave or folder, respectively.
//							Using the quick profile menu automatically selects the outputted profiles in the data browser.
//							Added z-rotation controls for 2D image waves.
//							Upgraded code to Igor 8.
//							Various bug-fixes and performance improvements.
//							Logs are now printed by default by the UI.
// 2023-12-07 - ver. 4.25:	Added TicksFromWaveLabels() function to generate user ticks from dimension labels.
//							Fixed bug: If a sub-range was used for alignment then the image data was cut to this sub-range.
//							Added keyboard shortcut: Hold 'alt / option' to save output in folder of the source wave.
//							Added feature to display profiles in z direction.
//							Fixed bug: Column units of multi-profiles were not set.
//							Fixed bug: Quit panel had the wrong size with active graph expansion (i.e., different from one).
//							Fixed bug: 3D profile calculation was wrong if the ROI was set to the minimal width.
//							CombineDataSets() now returns the name of the combined wave.
//							CombineDataSets() was renamed to QuickCombineData() and works for all three dimensions now.
// 2024-03-05 - ver. 4.30:	Optimized handling code for zProfile subwindow.
//							Fixed bug: Quick-combine output names had two underscore characters.
//							'Save as' now asks the user if existing waves should be overwritten.
//							Added checkbox to toggle to a log-scale plot.
// 2024-06-16 - ver. 4.35:	Fixed bug: Two 3D waves could not be stacked.
// 							Fixed bug in the UI: It was possible to set negative widths via the keyboard.
//							Fixed UI error with older panels.
//							Updated rotation button labels to look good on both Mac and Windows.
// 2024-08-03 - ver. 4.36:	Fixed bug: Undoing a single align action resulted in temporary waves in the wrong location.
//							QuickCombine() now adds all wave notes to the combined wave.
//							QuickCombine() was not outputting the correct size when combining waves with different, non-overlapping ranges.
// 2024-08-09 - ver. 4.40:	Added toggle to fill gaps created by the align feature.
//							Added scaling to the profile subtraction feature.
//							Fixed a crash with Igor 8.
// 2024-12-02 - ver. 4.45:	Fixed error when switching modes while unrelated traces where on the graph.
//							The trace separation of multiple profiles is now calculated in a more robust way from the data maximum.
//							Now an external wave can be loaded as a profile for comparison, division and subtraction.
//							The smoothing for normalization can now be set directly in the panel.
//							Normalization smoothing is also applied for maximum / area normalization.
//							Added area normalization feature.
//							Added format string for wave naming to QImageSplit(). Use the shift key to invoke from the menu.
//________________________________________________________________________________________________


//-------------------------------- functions for user code ---------------------------------------
static StrConstant kUserInfoBtnName = "Current"													// the title of the user button

static function Profile_ActivateUserInfoButton(wave inwave)										// called to check whether or not the data fulfills conditions for measurement normalization
	int activate = 0
	// +++++++++++++++++++++++++++ conditions for activation +++++++++++++++++++++++++++++++++++++
	activate = StringMatch(GetDimLabel(inwave,1,0),"*mA*")										// check if dim-labels contain "mA" values
	activate = StringMatch(note(inwave),"*ring current normalized*")? 0 : activate				// already normalized?
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return activate																				// return 1 to activate button
end

static function Profile_ExecuteUserInfoButton(wave inwave)										// normalization of the data to measurement info (when the button is pressed)
	variable averageValue = 0
	// ++++++++++++++++++++++ ring-current normalization procedure +++++++++++++++++++++++++++++++
	int i, dim = WaveDims(inwave)
	variable value
	for (i = 0; i < DimSize(inwave,dim-1); i++)
		string curLabel = GetDimLabel(inwave,1,i)												// get the information from the column label
		sscanf curLabel, "%s (%f mA)", curLabel, value											// search for ring current info in the form "time (XXX mA)"
		if (value != 0 && numtype(value) == 0)
			if (dim == 3)
				inwave[][][i] /= value
			else
				inwave[][i] /= value
			endif
			averageValue += value																// add the values for intensity re-normalization later
		endif
	endfor
	averageValue /= i+1
	Note inwave, "ring current normalized"														// write into note that the normalization has been done
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return averageValue																			// return the intensity normalization factor which may be used restore the average height
end
//------------------------------------------------------------------------------------------------

static StrConstant kWorkingDir	 = "Profile_Generator"											// the name for the working folder
static StrConstant kSettingsFile = "Spectra Tools settings.dat"									// just for keyboard shortcuts for now
static StrConstant kMultiProfileColor = "PastelsMap"											// color map name for multi-profile coloring
static Constant	kNormSmooth	 = 2			// OBSOLETE / backwards compatibility - profile normalization: integer factor for smoothing the profile before division (0 = no smoothing)
static Constant	kAlignSmooth = 2			// OBSOLETE / backwards compatibility - align features: smoothing factor in percent of the wave's x/y dimension (e.g., 2 => points * 0.02)
static Constant kMaxMultiProfiles  = 201	// upper limit for the number of simultaneously displayed profiles of 3D volumes
static Constant kShowLabelDisplay  = 1		// enable to show dimension-label annotation
static Constant kEnableLogPrintout = 1		// enable to print modification log by default

Menu "Spectra Tools"
	SpectraImageProfile#MenuEntry("start_GUI")	,/Q, CreateBrowser; SpectraProfile($GetBrowserSelection(0))
	"-"
	SpectraImageProfile#MenuEntry("append_2D")	,/Q, CreateBrowser; QuickCombineData(1,SpectraImageProfile#selectionList())
	SpectraImageProfile#MenuEntry("stack_2D")	,/Q, CreateBrowser; QuickCombineData(2,SpectraImageProfile#selectionList())
	SpectraImageProfile#MenuEntry("bin_data")	,/Q, CreateBrowser; QuickDataBinWrapper()
	"-"
	SpectraImageProfile#MenuEntry("profile_H")	,/Q, CreateBrowser; QuickMenuWrapper(1,"QProfile")
	SpectraImageProfile#MenuEntry("profile_V")	,/Q, CreateBrowser; QuickMenuWrapper(0,"QProfile")
	SpectraImageProfile#MenuEntry("profile_L")	,/Q, CreateBrowser; QuickMenuWrapper(2,"QProfile")
	"-"
	SpectraImageProfile#MenuEntry("split_H")	,/Q, CreateBrowser; QuickMenuWrapper(1,"QImageSplit")
	SpectraImageProfile#MenuEntry("split_V")	,/Q, CreateBrowser; QuickMenuWrapper(0,"QImageSplit")
	SpectraImageProfile#MenuEntry("split_L")	,/Q, CreateBrowser; QuickMenuWrapper(2,"QImageSplit")
End

#if (IgorVersion() >= 9.0)
Menu "DataBrowserObjectsPopup"
	"Append 1D,2D"				,/Q, QuickCombineData(1,SpectraImageProfile#selectionList())
	"Stack 2D->3D"				,/Q, QuickCombineData(2,SpectraImageProfile#selectionList())
	"Average 2D,3D"				,/Q, QuickDataBinWrapper()
	"-"
	"Profile - [Horizontal]"	,/Q, QuickMenuWrapper(1,"QProfile")
	"Profile - [Vertical]"		,/Q, QuickMenuWrapper(0,"QProfile")
	"Merge 3D - [Layers]"		,/Q, QuickMenuWrapper(2,"QProfile")
	"-"
	"Split Data - [Horizontal]"	,/Q, QuickMenuWrapper(1,"QImageSplit")
	"Split Data - [Vertical]"	,/Q, QuickMenuWrapper(0,"QImageSplit")
	"Split Data - [Layers]"		,/Q, QuickMenuWrapper(2,"QImageSplit")
	"-"
End
#endif

//################################################################################################

static function/S MenuEntry(string which)
	string menuStr = "", searchStr = ""
	StrSwitch (which)
		case "start_GUI":
			menuStr = "Image Profile ..."
			searchStr = "start image profile"
		break
		case "append_2D":
			menuStr = "Append 1D,2D Datasets into 2D Data"
			searchStr = "append 2D data"
		break
		case "stack_2D":
			menuStr = "Stack 2D Datasets into 3D Data"
			searchStr = "stack 2D data"
		break
		case "bin_data":
			menuStr = "Average 2D|3D Data (Binning)"
			searchStr = "quick data binning"
		break
		case "profile_H":
			menuStr = "Profile of 2D|3D Data - [Horizontal]"
			searchStr = "quick horizontal profile"
		break
		case "profile_V":
			menuStr = "Profile of 2D|3D Data - [Vertical]"
			searchStr = "quick vertical profile"
		break
		case "profile_L":
			menuStr = "Image from 3D Data - [Layers]"
			searchStr = "quick layer profile"
		break
		case "split_H":
			menuStr = "Split Up 2D|3D Data - [Horizontal]"
			searchStr = "data split horizontal"
		break
		case "split_V":
			menuStr = "Split Up 2D|3D Data - [Vertical]"
			searchStr = "data split vertical"
		break
		case "split_L":
			menuStr = "Split Up 3D Data - [Layers]"
			searchStr = "data split layers"
		break
		default:
			return ""
	EndSwitch
	
	int fileID; string read = ""
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFile
	if (!V_flag)
		do
			FReadLine fileID, read
		while (strlen(read) > 0 && !StringMatch(read, searchStr+"*"))
		read = cleanStr(read,"\t| |")
		read = StringByKey(cleanStr(searchStr,"\t| |"),read,"=","\r")
		Close fileID
	endif
	return menuStr+SelectString(strlen(read),"","/"+read)
end

//################################################################################################

function QuickMenuWrapper(int dim, string type)
	string wName = "", select = "", selList = selectionList()
	int i, shftKey = ((GetkeyState(0) & 4)!=0)													// when shift is pressed ask for data range
	StrSwitch (type)
		case "QProfile":
			variable mode=1, P1=0, P2=inf
			if (shftKey)
				Prompt mode, "the range is in ...", popup "scaled values;data points;"
				Prompt P1, "from:"
				Prompt P2, "to:"
				DoPrompt "Integration Over a Limited Range.", mode, P1, P2
				if (V_flag)
					return -1
				endif
			endif
			
			for (i=0; i<ItemsInList(selList); i++)
				wave data = $StringFromList(i,selList)
				if (shftKey)
					if(mode==1)
						wName = QProfile(data,dim,from=P1,to=P2)
					else
						wName = QProfile(data,dim,fromPoint=P1,toPoint=P2)
					endif
				else
					wName = QProfile(data,dim)
				endif
				if (strlen(wName))
					select += GetWavesDataFolder(data,1) + wName + ";"
				endif
			endfor
			
			if (strlen(select))																	// selects the newly created waves
				ModifyBrowser clearSelection, selectList=select
			endif
		break
		case "QImageSplit":
			string formatStr = StrVarOrDefault("root:DataSplitFormat",defaultSplitNameFormat)
			if (shftKey)
				Prompt formatStr, "Format: (name = input wave name,   position = current scaled position,   xxx = counter digits,   Ddd = dimension letters in upper or lower case)"
				DoPrompt "Select naming scheme for individual wave.", formatStr
				if (V_flag)
					return -1
				endif
				string/G root:DataSplitFormat = formatStr
			endif
			for (i=0; i<ItemsInList(selList); i++)
				QImageSplit($StringFromList(i,selList),dim,format=formatStr)
			endfor
		break
	EndSwitch
	return 0
end

//________________________________________________________________________________________________
//	Quick Profile - averages 2D/3D waves in one direction (along x, y or z) for data reduction
//________________________________________________________________________________________________

function/S QProfile(wave/Z inwave, int mode [,variable from,variable to,variable fromPoint,variable toPoint,int quiet])
	if (!WaveExists(inwave))
		return ""
	endif
	int alert = ParamIsDefault(quiet) ? 1 : quiet == 0
	int wdim  = WaveDims(inwave)
	if (checkMode(mode, wdim, alert))
		return ""
	endif
	
	string name = NameOfWave(inwave)
	string outName = name + SelectString(mode-1, "_prV", "_prH","_prL")
	variable tmpP, P1 = 0, P2 = DimSize(inwave,mode)-1
	if (!ParamIsDefault(from) || !ParamIsDefault(to) || !ParamIsDefault(fromPoint) || !ParamIsDefault(toPoint))
		P1 = !ParamIsDefault(fromPoint) ? limit(fromPoint, 0, DimSize(inwave,mode)-1) : P1
		P2 = !ParamIsDefault(toPoint)   ? limit(toPoint,   0, DimSize(inwave,mode)-1) : P2
		P1 = !ParamIsDefault(from) && numtype(from) == 0 ? limit(ScaleToIndex(inwave, from, mode ), 0, DimSize(inwave,mode)-1)  : P1
		P2 = !ParamIsDefault(to)   && numtype(to)   == 0 ? limit(ScaleToIndex(inwave, to,   mode ), 0, DimSize(inwave,mode)-1)  : P2
		if (P1 > P2)
			tmpP = P1
			P1 = P2
			P2 = tmpP
		endif
		if     (mode == vertiMode)
			Duplicate/Free/RMD=[P1,P2][][] inwave, tempwave
		elseif (mode == horizMode)
			Duplicate/Free/RMD=[][P1,P2][] inwave, tempwave
		elseif (mode == layerMode)
			Duplicate/Free/RMD=[][][P1,P2] inwave, tempwave
		endif
		wave inwave = tempwave
		outName = name + SelectString(mode-1, "_V[", "_H[", "_L[") + num2str(P1)+"-"+num2str(P2) + "]"
	endif
	
	if     (mode == vertiMode)
		MatrixOP/O $outName = sumRows(inwave^t)/numRows(inwave)
	elseif (mode == horizMode)
		MatrixOP/O $outName = sumRows(inwave)/numCols(inwave)
	elseif (mode == layerMode)
		MatrixOP/O $outName = sumBeams(inwave)/waveLayers(inwave)
	endif
	wave outwave = $outName
	
	if (wdim == 3 && mode < layerMode)															// for profiles of 3D waves -> rotate result to 2D map
		MatrixOP/Free tempVol = transposeVol(outwave,1)
		Duplicate/O tempVol, outwave
		SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), outwave
	endif
	
	variable X1 = DimOffset(inwave,mode)
	variable X2 = DimOffset(inwave,mode)+(DimSize(inwave,mode)-1)*DimDelta(inwave,mode)
	string unit = WaveUnits(inwave,mode);	unit = SelectString(strlen(unit),""," ")+unit
	string notes = note(inwave) + "\r" + SelectString(mode-1,"Rows","Columns","Layers") + " of " + name
	notes += " integrated from "+num2str(X1)+" to "+num2str(X2)
	notes +=   unit +" (points "+num2str(P1)+" to "+num2str(P2)+")"
	Note outwave, notes
	
	if (mode < layerMode)
		SetScale/P x, DimOffset(inwave,1-mode), DimDelta(inwave,1-mode), WaveUnits(inwave,1-mode), outwave
	elseif (mode == layerMode)
		CopyScales inwave, outwave
	endif
	SetScale d 0,0, WaveUnits(inwave,-1), outwave												// also copy the data units
	
	return outName
end

//________________________________________________________________________________________________
//	Image split - splits 2D/3D data into individual waves along x, y or z
//________________________________________________________________________________________________

static StrConstant defaultSplitNameFormat = "ddd_xxx_(position)"

// format string explanation:
// name		- will be replaced with the wave name
// position	- will be replaced with the current data position
// dd / DD	- dimension indicator like 'row' or 'column'; letter count / case decide length / font case
// xxx		- split counter; letter count decides length

function/S QImageSplit(wave/Z inwave, int mode [, int quiet, string format])
	if (!WaveExists(inwave))
		return ""
	endif
	int alert = ParamIsDefault(quiet) ? 1 : quiet == 0
	int wdim  = WaveDims(inwave)
	if (checkMode(mode, wdim, alert))
		return ""
	endif
	
	string formatStr = defaultSplitNameFormat
	if (!ParamIsDefault(format) && strlen(format))
		formatStr = SelectString(StringMatch(format,"*x*")||StringMatch(format,"*POSITION*"),format+"_X",format)	// make sure there is at least one digit
	endif
	
	string wName = NameOfWave(inwave), dimFormat, numFormat
	string wUnit = WaveUnits(inwave,mode);	wUnit = SelectString(strlen(wUnit),""," "+wUnit)
	string notes = note(inwave);			notes = SelectString(strlen(notes),"",notes+"\r")
	
	DFREF saveDFR = GetDataFolderDFR()
	NewDataFolder/O/S $wName
	variable i, j																				// must be variable for MatrixOP
	for (i = 0; i < DimSize(inwave,mode); i++)
		string dimStr = SelectString(mode-1,"row","column","layout"), tmp = ""
		string posStr = num2str(DimOffset(inwave,mode)+i*DimDelta(inwave,mode)) + wUnit
		SplitString/E=("[^dD]*([dD]*)") formatStr, dimFormat
		SplitString/E=("[^xX]*([xX]*)") formatStr, numFormat
		string outName = formatStr
		if (strlen(dimFormat))
			for (j = 0; j < min(strlen(dimFormat),strlen(dimStr)); j++)
				if (!CmpStr(dimFormat[j],"D",1))
					tmp += UpperStr(dimStr[j])
				else
					tmp += dimStr[j]
				endif
			endfor
			tmp += SelectString(wdim==3 && mode < layerMode,"","_img")
			outName = ReplaceString(dimFormat,outName,tmp)
		endif
		if (strlen(numFormat))
			sprintf tmp, "%.*d", strlen(numFormat), i
			outName = ReplaceString(numFormat,outName,tmp)
		endif
		outName = ReplaceString("NAME",ReplaceString("POSITION",outName,posStr),wName)
		
		if (wdim == 3)
			if     (mode == vertiMode)
				MatrixOP/Free tempData = row(inWave,i)
				MatrixOP/O $outName = transposeVol(tempData,4)
			elseif (mode == horizMode)
				MatrixOP/Free tempData = col(inWave,i)
				MatrixOP/O $outName = transposeVol(tempData,1)
			elseif (mode == layerMode)
				MatrixOP/O $outName = inWave[][][i]
			endif
		else
			if     (mode == vertiMode)
				MatrixOP/O $outName = row(inWave,i)^t
			elseif (mode == horizMode)
				MatrixOP/O $outName = col(inWave,i)
			endif
		endif
		
		int xDim =   (mode == vertiMode)		// always zero unless vertical mode
		int yDim = 2-(mode == layerMode)		// always two unless layer mode
		SetScale/P x, DimOffset(inwave,xDim), DimDelta(inwave,xDim), WaveUnits(inwave,xDim), $outName
	if (wdim == 3)
		SetScale/P y, DimOffset(inwave,yDim), DimDelta(inwave,yDim), WaveUnits(inwave,yDim), $outName
	endif
		SetScale d 0,0, WaveUnits(inwave,-1), $outName
		Note $outName, notes + SelectString(mode-1,"row","column","layer") + " " + num2str(i) + " of " + NameOfWave(inwave)
	endfor
	SetDataFolder saveDFR
	
	return wName
end

//________________________________________________________________________________________________
//	Combines individual 1D or 2D datasets to one 2D/3D dataset
//________________________________________________________________________________________________

function/S QuickCombineData(int mode, string list)												// combines the 'list' of waves in x(0), y(1) or z(2) directions
	int i, j, items = ItemsInList(list), layers
	if (items < 2 || mode < 0 || mode > 2)
		return ""
	endif
	
	// +++++++++++++++++++++++++++++++++ build an output name ++++++++++++++++++++++++++++++++++++
	string number1, number2, name, outname
	name = ParseFilePath(0, StringFromList(items-1, list), ":", 1, 0)							// try to get the last data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number2													// get the end number in the name
	name = ParseFilePath(0, StringFromList(0, list), ":", 1, 0)									// try to get the first data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number1
	
	outname = RemoveEnding(cleanStr(name,"'"), number1)
	outname = RemoveEnding(outname, "_")
	if(strlen(number1) > 0 && strlen(number2) > 0)												// add range to name if both numbers are set
		outname += "_"+number1+"-"+number2
	endif
	outname = outname + "_" + SelectString(mode-1, "cbV", "cbH","cbL")
	
	// ++++++++++++++++++++++++++++++++++ check data sizes +++++++++++++++++++++++++++++++++++++++
	Make/FREE/N=(items) mainN = DimSize($StringFromList(p,list),1-(mode>0))
	Make/FREE/N=(items) subdN = DimSize($StringFromList(p,list),(mode>0))
	mainN = mainN[p] == 0 ? 1: mainN[p]
	subdN = subdN[p] == 0 ? 1: subdN[p]
	
	if (mode == 2)
		if (mainN[0]!=mean(mainN) || subdN[0]!=mean(subdN))										// don't support different data sizes
			Print "Combine error: X- and Y-dimensions must match for layer stacking."
			return ""
		endif
		Concatenate/O/NP=2 list, $outname
		string namelist = ""
		for (i = 0; i < items; i++)																// first, create a list of labels for each layer
			layers = DimSize($StringFromList(i,list),2)
			layers = layers == 0 ? 1: layers
			for (j = 0; j < layers; j++)
				namelist += ParseFilePath(0, StringFromList(i, list), ":", 1, 0)+";"
			endfor
		endfor
		for (i = 0; i < DimSize($outname,2); i++)
			name = cleanStr(NameOfWave($StringFromList(i, namelist)),"\"|'|,|;|")[0,255]
			SetDimLabel 2, i, $name, $outname													// label each layer with their input
		endfor
		return outname																			// stop here
	endif
	
	// ++++++++++++++++++++++++++++++ continue for modes 0 and 1 +++++++++++++++++++++++++++++++++
	Make/FREE/N=(items) Offset = DimOffset($StringFromList(p,list),1-mode)						// temporary data holding all the scaling and unit values from all waves	
	Make/FREE/N=(items) Deltas = DimDelta( $StringFromList(p,list),1-mode)
	Make/FREE/N=(items) RngEnd = Offset[p] + Deltas[p] * (mainN[p]-1)
	
	variable Delta = Deltas[0], deltaMismatch = (Delta!=mean(Deltas))
	variable stackN = mainN[0], mainNMismatch = (stackN!=mean(mainN))
	if (deltaMismatch)																			// may lead to data overrun
		DoAlert 1, "Data x spacing does not match! Do you want to use interpolation? Press no to exit."
		if (V_flag != 1)
			return ""
		endif
		Delta = Delta > 0 ? WaveMin(Deltas) : WaveMax(Deltas)
	endif
	variable Y1 = 0, Y2
	variable X1	= Delta > 0? WaveMin(Offset) : WaveMax(Offset), offsetMismatch = (X1!=mean(Offset))
	variable X2	= Delta > 0? WaveMax(RngEnd) : WaveMin(RngEnd), rngEndMismatch = (X2!=mean(RngEnd))
	if (deltaMismatch || mainNMismatch || offsetMismatch || rngEndMismatch)
		stackN = abs((X1-X2)/Delta)+1															// recalculate point size
	endif
	
	wave/Z work = $StringFromList(0,list)														// extract scaling for opposite dimension from first wave only
	if (!WaveExists(work))
		Print "Combine error: Some of the input waves do not exist."
		return ""
	endif
	
	if (mode)
		Make/O/D/N=(stackN,Sum(subdN)) $outname/wave=out = 0
		SetScale/I x, X1, X2, WaveUnits(work,0), out
		SetScale/P y, DimOffset(work,1), DimDelta(work,1), SelectString(WaveDims(work)>1,"",WaveUnits(work,1)), out
	else
		Make/O/D/N=(Sum(subdN),stackN) $outname/wave=out = 0
		SetScale/P x, DimOffset(work,0), DimDelta(work,0), WaveUnits(work,0), out
		SetScale/I y, X1, X2, WaveUnits(work,1), out
	endif
	SetScale d 0,0, WaveUnits(work,-1), out
	
	string wnotes = "", cmbNote = "", header = "Combined data of "
	for (i = 0; i < ItemsInList(list); i++)
		wave work = $StringFromList(i, list)
		if (!WaveExists(work))
			Print "Combine error: '"+StringFromList(i, list)+"'does not exist."
			continue
		endif
		
		header += NameOfWave(work) + ", "
		wnotes += SelectString(strlen(note(work)),"","\r\r----- notes of "+NameOfWave(work)+" -----\r" + note(work))
		
		name = cleanStr(NameOfWave(work),"\"|'|,|;|")[0,255]
		for (j = 0; j < subdN[i]; j += 1)
			SetDimLabel mode, Y1+j, $name, out													// label added columns with the input wave
		endfor
		
		X1  = ScaleToIndex(out,IndexToScale(work, 0			,1-mode),1-mode)					// find the valid sub-range from the data
		X2  = ScaleToIndex(out,IndexToScale(work, mainN[i]-1,1-mode),1-mode)
		X1 += (IndexToScale(work, 0,1-mode) - IndexToScale(out, X1		  ,1-mode)) >= Delta	// make sure the x-range is not outside the valid range
		X2 -= (IndexToScale(out, X2,1-mode) - IndexToScale(work,mainN[i]-1,1-mode)) >= Delta
		Y2  = Y1 + subdN[i]-1																	// next stack
		if (mode)
			out[X1,X2][Y1,Y2] = work(x)[q-Y1]
		else
			out[Y1,Y2][X1,X2] = work[p-Y1](y)
		endif
		Y1 += subdN[i]
	endfor
	Note out, RemoveEnding(header, ", ")+wnotes
	return outname
end

//________________________________________________________________________________________________
//	Bin data in x, y, z
//________________________________________________________________________________________________

function QuickDataBinWrapper()
	string list = selectionList()
	int i, items = ItemsInList(list)
	if (!items)
		return -1
	endif
	
	variable binX = NumVarOrDefault("root:DataBin_X",1)
	variable binY = NumVarOrDefault("root:DataBin_Y",1)
	variable binZ = NumVarOrDefault("root:DataBin_Z",1)
	Prompt binX, "X dimension:"
	Prompt binY, "Y dimension:"
	Prompt binZ, "Z dimension:"
	string helptext = "Creates a lower-resolution representation of the data by using ImageInterpolate."
	DoPrompt/HELP=helptext "Enter how many points should be averaged in each dimension", binX, binY, binZ
	
	for (i = 0; i < items; i++)
		QuickBinData($StringFromList(i, list), binX, binY, binZ)
	endfor
	
	variable/G root:DataBin_X = binX															// save for later use
	variable/G root:DataBin_Y = binY
	variable/G root:DataBin_Z = binZ
	return 0
end

//################################################################################################

function QuickBinData(wave in, variable binX, variable binY, variable binZ)
	if (WaveDims(in) < 2)																		// only for images
		return 0
	endif
	binX = binX < 1 ? 1 : round(binX)
	binY = binY < 1 ? 1 : round(binY) 
	binZ = binZ < 1 ? WaveDims(in) == 3 : round(binZ)*(WaveDims(in) == 3)
	ImageInterpolate/PXSZ={binX,binY,binZ}/DEST=$(NameOfWave(in)+"_avg") Pixelate in			// average image
	wave out = $(NameOfWave(in)+"_avg")															// transfer scales and notes
	SetScale/P x DimOffset(in,0), DimDelta(in,0)*binX, WaveUnits(in,0), out
	SetScale/P y DimOffset(in,1), DimDelta(in,1)*binY, WaveUnits(in,1), out
	SetScale/P z DimOffset(in,2), DimDelta(in,2)*binZ, WaveUnits(in,2), out
	string paraStr = ""																			// write notes
	paraStr += SelectString(binX != 1,"", "x="+num2str(binX))
	paraStr += SelectString(strlen(paraStr),"", ", ")
	paraStr += SelectString(binY != 1,"", "y="+num2str(binY))
	paraStr += SelectString(strlen(paraStr),"", ", ")
	paraStr += SelectString(binZ != 1,"", "z="+num2str(binZ))
	if (strlen(paraStr))
		Note/K out, note(in) + "\r"+NameOfWave(in)+" averaged with a bin size of "+paraStr+"."
	endif
	return 0
end

//################################################################################################

static function/S cleanStr(string str, string list [, string insert])
	int i; string replace = SelectString(ParamIsDefault(insert),insert,"")
	for (i = 0; i < ItemsInList(list,"|"); i++)
		str = ReplaceString(StringFromList(i,list,"|"), str, replace)
	endfor
	return str
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S selectionList()
	string out = "", curr = ""
	int i = 0
	do
		curr = GetBrowserSelection(i++)
		if (Exists(curr) == 1)
			out += curr + ";"
		endif
	while(strlen(curr) > 0)
	return Sortlist(out, ";", 16)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Constant vertiMode = 0
static Constant horizMode = 1
static Constant layerMode = 2

static function checkMode(int mode, int wdim, int alert)
	if (mode < vertiMode || mode > layerMode)
		if (alert)
			DoAlert 0, "Wrong dimension parameter. Must be 0, 1 or 2"
		endif
		return 1
	endif
	if (wdim < 2)
		if (alert)
			DoAlert 0, "This is meant for 2D or 3D waves!"
		endif
		return 1
	endif
	if (wdim != 3 && mode == layerMode)
		if (alert)
			DoAlert 0, "This is meant for 3D waves!"
		endif
		return 1
	endif
	return 0
end

//################################################################################################

function TicksFromWaveLabels(wave in, int dim)													// creates label waves to plot contained dimension labels together with the input wave in a graph
	variable i,size = DimSize(in,dim)
	variable offset = DimOffset(in,dim), delta = DimDelta(in,dim)
	if (size < 2)
		return -1
	endif
	Make/Free/T/N=(size) full_labels = GetDimLabel(in, 1, p)
	Make/Free/D/N=(size) index = p
	FindDuplicates/Free/INDX=duplicate_id/RT=labels_clean full_labels
	for (i=0; i<DimSize(duplicate_id,0); i++)
		index[duplicate_id[i]] = NaN
	endfor
	WaveTransform zapNaNs, index
	index[DimSize(index,0)] = {DimSize(in,dim)-1}				// add end of wave
	
	Make/O/T/N=(DimSize(labels_clean,0)*2+1,2) $(NameOfWave(in)+"_tickTxt")/Wave=labels=""
	Make/O/D/N=(DimSize(labels_clean,0)*2+1)   $(NameOfWave(in)+"_tickVal")/Wave=values=NaN
	SetDimLabel 1, 1, $("Tick Type"), labels
	for (i=0;; i++)
		labels[i*2][0]	 = ""
		labels[i*2][1]	 = "Major"
		values[i*2]		 = offset+delta*index[i]
		if (i >= DimSize(labels_clean,0))
			break
		endif
		labels[i*2+1][0] = SelectString(strlen(labels_clean[i]), "none", labels_clean[i])
		labels[i*2+1][1] = "Minor"
		values[i*2+1]	 = offset+delta*((index[i+1]-index[i])/2+index[i])
	endfor
end

//________________________________________________________________________________________________
//	Generate profiles from 2D or 3D waves
//________________________________________________________________________________________________

function SpectraProfile(wave/Z inwave)
	if (!WaveExists(inwave))
		return -1
	endif
	int dim = WaveDims(inwave), lnum = DimSize(inwave,2)
	if (dim < 2 || dim > 3 || WaveType(inwave,1) != 1 )
		Abort "This tool works only for numeric 2D or 3D waves."
	endif
	
	//+++++++++++++++++++++++++++++++++ generate working variables ++++++++++++++++++++++++++++++
	string gTitle = "Spectra Image Profile (ver. "+num2str(kVersion)+" - "+kVersionDate+"): Profile of "+NameOfWave(inwave)
	string gName = UniqueName("ProfileGeneratorGraph", 6, 1)
	string pName = "ImageProfilePanel"
	string workF = "IP"+ReplaceString("ProfileGeneratorGraph",gName,"")+"_"+CleanupName(NameOfWave(inwave),0)[0,25]
	
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:$(kWorkingDir)
	NewDataFolder/O root:Packages:$(kWorkingDir):$(workF)
	NewDataFolder/O	root:WinGlobals
	NewDataFolder/O	root:WinGlobals:$(gName)													// here all main window related data is generated
	DFREF stg = root:Packages:$(kWorkingDir):$(workF)
	string/G	root:WinGlobals:$(gName):S_TraceOffsetInfo = ""									// Igor defined window status parameter
	string/G			stg:InwaveLocation = GetWavesDataFolder(inwave,2)						// save wave name and folder for later
	variable/G			stg:ProfileUpdateConnector												// formula connection variable... has no purpose itself, but is an anchor for triggering updates
	NVar updateVar	  = stg:ProfileUpdateConnector
	Make/O/D/N=2		stg:SectionLine															// the currently displayed profile
	Make/O/D/N=(1,lnum)	stg:MultiSection														// for multi-profiles of 3D volumes
	Make/O/D/N=5		stg:CurrentLines = 0													// holds the range values ([0] = current working mode {0 = horizontal, 1 = vertical}, [1] = h.pos, [2] = h.width, [3] = v.pos, [4] = v.width)
	Make/O/D/N=5 		stg:LinesY/wave=LinesY, stg:LinesX/wave=LinesX							// marquee line waves for the image 
	Duplicate/O	inwave, stg:StartCopy															// copy original wave for undo
	SetScale d 0,0,WaveUnits(inwave,0), LinesX													// the data units will be displayed on the x- and y-axes
	SetScale d 0,0,WaveUnits(inwave,1), LinesY

	//+++++++++++++++++++++++++++++++++ generate the graph ++++++++++++++++++++++++++++++++++++++
	variable TFTResX, TFTResY, pix = 72/ScreenResolution										// get the display resolution to align window better (center of full screen)
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,","))-Str2Num(StringFromList(0,TFTInfo,",")))*pix
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,","))-Str2Num(StringFromList(1,TFTInfo,",")))*pix
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	variable wT = 70, wL = 250, pW = 270, pH = 620 + 110*(dim == 3)
	variable wB	= limit(TFTResY- 70, wT+300, inf)
	variable wR	= limit(TFTResX-350, wL+300, inf)
	
	Display/K=2/W=(wL*pix,wT*pix,wR*pix,wB*pix)/N=$gName as gTitle
	AppendImage/G=1 inwave																		// don't append as RGB image
	ModifyImage ''#0 ctab= {*,*,Terrain,0},	ctabAutoscale=2										// only calculate autoscale from current layer
	AppendToGraph/C=(0,0,0) LinesY vs LinesX													// add marquee lines
	ModifyGraph quickdrag(LinesY)=1,live(LinesY)=1,rgb(LinesY)=(65535,0,0)						// lines can be dragged (act on S_TraceOffsetInfo)
	SetWindow $gName hook(CreateProfile)=Profile_ProcessWindowEvent								// keeps track of window changes
	SetFormula updateVar,"LinesIntegrateUpdate(root:WinGlobals:"+gName+":S_TraceOffsetInfo)"	// set formula for live update
	
	//++++++++++++++++++++++++++++++++++++ now the controls ++++++++++++++++++++++++++++++++++++++
	NewPanel/HOST=$gName/EXT=0/K=2/N=$pName/W=(0,0,pW,pH) as "Spectra Profile"
	pName = gName+"#"+pName
	SetWindow $pName hook(CreateProfile)=Profile_ProcessWindowEvent								// keeps track of window changes here as well (only button presses)
	SetWindow $pName userdata(workFolder)=workF													// save work folder location
	SetWindow $gName userdata(procVersion)=num2str(kVersion)									// save procedure version
	
	TitleBox IntMode			,pos={13,7}		,size={80, 23}	,title="\Zr125Mode:"	 						,frame=0	,fstyle=1
	CheckBox IntColCheckBox		,pos={70,10}	,size={80, 23}	,title="\f05H\f01orizontal: ═"		,help={"Select horizontal mode. (key: 'h')"}
	CheckBox IntRowCheckBox		,pos={175,10}	,size={80, 23}	,title="\f05V\f01ertical: ║"		,help={"Select vertical mode. (key: 'v')"}
	SetVariable Position		,pos={10,33}	,size={155,23}	,title="Center:"					,help={"Center position of the selection. Move with cursor keys."}	
	SetVariable Width			,pos={10,56}	,size={155,23}	,title="Width:"						,help={"Width of the selection. This is the distance between the two lines. Change with cursor keys."}
	CheckBox FullWidth			,pos={175,34}	,size={70, 23}	,title="Full Width"					,help={"Select the full range of the spectrum. (key: page up)"}
	CheckBox MinWidth			,pos={175,57}	,size={70, 23}	,title="Min. Width"					,help={"Select one slice. (key: page down)"}
	
	Button bSaveSpec			,pos={20,84}	,size={100,28}	,title="\f04S\f00ave Profile"		,help={"Save current profile. (key: 's') Hold 'shift' to omit the profile range in the wave name. Hold 'alt / option' to save the output in the source's folder."}
	Button bCompare				,pos={20,120}	,size={100,23}	,title="Sna\f04p\f00shot"			,help={"Click to copy and compare current profile as a snapshot. (key: 'p')"}	,valueColor=(0,35000,0)
	Button LoadProfile			,pos={20,151}	,size={100,23}	,title="Load Profile"				,help={"Loads an external wave for comparison with the current profile."}		,valueColor=(34000,10000,60000)

	CheckBox SnapToGrid			,pos={135,83}	,size={80, 23}	,title="Snap Lines to \f04G\f00rid"	,help={"Snap the selection lines to full rows/columns. (key: 'g')"}
	CheckBox Normto1			,pos={135,105}	,size={80, 23}	,title="\f04N\f00ormalize Profile"	,help={"Normalizes all profiles to 1. (key: 'n')"}
	CheckBox MousetoSpec		,pos={135,128}	,size={80, 23}	,title="\f04M\f00ouse Csr Profile"	,help={"Shows a live profile below the mouse cursor. (key: 'm')"}
	CheckBox cPrintEdits		,pos={135,155}	,size={250,23}	,title="Print Log to History"								,fstyle=2
	
	GroupBox NormalizeGroup		,pos={10,183}	,size={250,86}	,title="\\Zr110Normalize Image / Active Layer"	,frame=0	,fstyle=1
	TitleBox NormModeTitle		,pos={22,211}	,size={80, 23}	,title="Each Col.:"	 							,frame=0
	Button bNormMax				,pos={79,207}	,size={42, 23}	,title="Max."						,help={"Normalize the image or selected layer to the maximum of every horizontal / vertical line. Holding 'ctrl / cmd' disables smoothing (hard normalize)."}
	Button bNormArea			,pos={128,207}	,size={42, 23}	,title="Area"						,help={"Normalize the image or selected layer to the area of every horizontal / vertical line. Holding 'ctrl / cmd' disables smoothing (hard normalize)."}
	Button bNormalize			,pos={20,237}	,size={70, 23}	,title="1 / Profile"				,help={"Normalize the image or selected layer by the current profile (division). Holding 'ctrl / cmd' disables smoothing (hard normalize)."}
	Button bNormUser			,pos={180,207}	,size={70, 23}	,title=kUserInfoBtnName				,help={"Normalize the image or selected layer with the included measurement information, e.g., ring current."}
	SetVariable NormSmooth		,pos={120,240}	,size={50, 23}	,title="Smth:"						,help={"Smooth the profile data before normalization (this helps with low S/N)."}
	CheckBox MaintainInt		,pos={180,241}	,size={80, 23}	,title="Keep Avg."					,help={"Check to maintain the (average) intensity of the data (scales all values back to the average)."}	
	
	GroupBox ModifyGroup		,pos={10,275}	,size={250,235}	,title="\\Zr110Edit Image / Volume"				,frame=0	,fstyle=1
	SetVariable SubtractScale	,pos={100,303}	,size={50, 23}	,title="x "							,help={"Scales the profile up or down for subtraction. Negative values will instead result in addition."}
	Button bSubtractPr			,pos={20,300}	,size={70, 23}	,title="Subtract"					,help={"Cut the selected profile from the image or current layer. Press and hold 'ctrl / cmd' (or 'alt / option') to subtract along features moving with increasing (decreasing) Y values."}
	Button bRemoveSelection		,pos={20,330}	,size={130,23}	,title="Delete ROI Section"			,help={"Remove the selected data between the lines from the image or ALL layers of 3D data. (key: del)"}
	Button bShiftYAxisPlus		,pos={160,300}	,size={90, 23}	,title="> Shift: +Y"				,help={"Shifts 2D / 3D data by adding Y-axis values to X values. Hold the 'ctrl / cmd' key, to cut image to exclude partial data."}
	Button bShiftYAxisMinus		,pos={160,330}	,size={90, 23}	,title="< Shift: -Y"				,help={"Shifts 2D / 3D data by subtracting Y-axis values from X values. Hold the 'ctrl / cmd' key, to cut image to exclude partial data."}
	
	GroupBox SeparationLine1	,pos={20,359}	,size={228, 1}	,frame=0
	Button bAlign				,pos={20,371}	,size={65, 55}	,title="Align\nFeatures"			,help={"Use data correlation to align features in the image or current layer (press 'shift' for top to bottom alignment). Set cursors to align to a sub-region!"}
	CheckBox cAlignSmooth		,pos={95,370}	,size={80, 23}	,title="Smoothing:"					,help={"Smooth profile before trying the alignment (this helps with low S/N)."}
	CheckBox cAlignOutput		,pos={95,412}	,size={80, 23}	,title="Save Result"				,help={"Saves the used shift value for each line alignment in a wave. This can be used to log the amount of shift or analyze this result further."}
	CheckBox cAlignSubShift		,pos={185,391}	,size={80, 23}	,title="Substeps"					,help={"Shifts data in sub-steps if possible (interpolation is used)."}
	CheckBox cAlignFillGaps		,pos={185,412}	,size={80, 23}	,title="Fill Gaps"					,help={"Fills the gaps created by shifting with the end points of each row/column. Only works with sub-step shifting enabled."}
	CheckBox cAlignEdges		,pos={95,391}	,size={80, 23}	,title="Align Edges"				,help={"Align edges instead of peak features (using the derivative)."}
	SetVariable SmoothFactor	,pos={140,368}	,size={110,23}	,title=""							,help={"The smoothing factor (no. of binomial smoothing operations) applied before alignment."}
	
	GroupBox SeparationLine2	,pos={20,435}	,size={228, 1}	,frame=0
	Button bUndoAll				,pos={20,445}	,size={90, 23}	,title="\f05U\f01ndo All"			,help={"Completely restore the original data. (key: 'u')"} ,fSize=13
	Button bUndoLast			,pos={20,475}	,size={90, 23}	,title="Undo Last"					,help={"Restore the data to the state before the last action."}
	Button bSaveQuit			,pos={120,445}	,size={130,23}	,title="Accept & Quit"				,help={"Keeps any changes made to the 2D data and closes the window."}
	Button bSaveAsNew			,pos={120,475}	,size={130,23}	,title="Save as ..."				,help={"Saves the edited 2D / 3D data under a new name. The original input will be restored."}
	
	CheckBox cDoLogScale		,pos={10,522}	,size={80, 23}	,title="Log Scale: "				,help={"Shows intensities on a log scale."}	,fStyle=1
//	TitleBox ContrastLabel		,pos={10,522}	,size={45, 23}	,title="Scale"									,frame=0	,fStyle=1
	TitleBox ContrLowLabel		,pos={10,542}	,size={24, 23}	,title="Min:"									,frame=0
	TitleBox ContrHighLabel		,pos={10,562}	,size={24, 23}	,title="Max:"									,frame=0
	
	PopupMenu ContrastColor		,pos={100,520}	,size={160,20}	,title=""			,value= "*COLORTABLEPOPNONAMES*"
	Slider ContrastLow			,pos={48,547}	,size={210,10}	,side=0	,vert=0	,ticks=0
	Slider ContrastHigh			,pos={48,567}	,size={210,10}	,side=0	,vert=0	,ticks=0
	
	if (dim == 3)
		GroupBox LayersGroup	,pos={10,585}	,size={250,137}	,title="\\Zr1103D Volume Controls"				,frame=0	,fstyle=1
		Button bDeleteLayer		,pos={120,608}	,size={50, 23}	,title="Delete"						,help={"Deletes the selected layer from the stack. This can be undone."}
		Button bSumLayers		,pos={180,608}	,size={70, 23}	,title="Merge All"					,help={"Merges layers of 3D data into one 2D image for processing. This can be undone."}
		SetVariable SetLayer	,pos={10,610}	,size={100,23}	,title="Layer:"						,help={"Selects the active layer of 3D data."}
	
		CheckBox MultiProfile	,pos={22,640}	,size={80, 23}	,title="Multiple Profiles"			,help={"When activated, profiles of all other layers are displayed as well."}
		SetVariable vMulti_Dist	,pos={150,638}	,size={100,23}	,title="Separation:"				,help={"Sets the vertical separation of multiple profiles in percent of the maximum y-value."}
		CheckBox doZProfiles	,pos={22,664}	,size={80, 23}	,title="Display \f04Z\f00-Profile:"	,help={"When activated, profiles in z-direction are displayed."}
		PopUpMenu doZProfMode	,pos={140,662}	,size={110,23}	,title=""							,help={"Selects which information is displayed in the profile."}
		
		TitleBox LayerRotMode	,pos={20,695}	,size={120,23}	,title="Rot:"									,frame=0
		Button bLayer_RotX_pos	,pos={48,690}	,size={23, 23}	,title="⤴"							,help={"Rotates the 3D volume around x (row) axis up."}						// U+2934
		TitleBox LayerRotX		,pos={75,695}	,size={20, 23}	,title="X"		 								,frame=0	,fstyle=1
		Button bLayer_RotX_neg	,pos={86,690}	,size={23, 23}	,title="⤵"							,help={"Rotates the 3D volume around x (row) axis down."}					// U+2935
		Button bLayer_RotY_pos	,pos={119,690}	,size={23, 23}	,title="⤶"							,help={"Rotates the 3D volume around y (column) axis to the left."}			// U+2936
		TitleBox LayerRotY		,pos={146,695}	,size={20, 23}	,title="Y"		 								,frame=0	,fstyle=1
		Button bLayer_RotY_neg	,pos={157,690}	,size={23, 23}	,title="⤷"							,help={"Rotates the 3D volume around y (column) axis to the right."}		// U+2937
		Button bLayer_RotZ_pos	,pos={188,690}	,size={23, 23}	,title="↺"							,help={"Rotates the image/3D volume around z (layer) axis to the left."}	// U+21ba
		TitleBox LayerRotZ		,pos={216,695}	,size={20, 23}	,title="Z"		 								,frame=0	,fstyle=1
		Button bLayer_RotZ_neg	,pos={227,690}	,size={23, 23}	,title="↻"							,help={"Rotates the image/3D volume around z (layer) axis to the right."}	// U+21bb
		
		CheckBox MultiProfile														,value = 0		,proc=ModeCheckBoxControl		,disable = 2*(lnum<2 || lnum>kMaxMultiProfiles)
		CheckBox doZProfiles														,value = 0		,proc=ModeCheckBoxControl		,disable = 2*(lnum<2)
		PopUpMenu doZProfMode	,bodywidth =110	,value=#("\""+kZProfileModes+"\"")					,proc=Profile_PopupControl		,mode=1
		SetVariable SetLayer	,bodywidth = 55	,limits={0,lnum-1,1}				,value =_NUM:0	,proc=ChangeParameterControl
		SetVariable vMulti_Dist	,bodywidth = 50	,limits={0,100,5}	,format="%g %"	,value =_NUM:10	,proc=ChangeParameterControl
	else
//		GroupBox LayersGroup	,pos={10,585}	,size={130,53}	,title="\\Zr110Transformation"					,frame=0	,fstyle=1
		TitleBox LayerRotMode	,pos={10,590}	,size={120,23}	,title="Rotate Image:"							,frame=0	,fStyle=1
		Button bLayer_RotZ_pos	,pos={100,587}	,size={23, 23}	,title="⟲"							,help={"Rotates the image to the left."}
		Button bLayer_RotZ_neg	,pos={132,587}	,size={23, 23}	,title="⟳"							,help={"Rotates the image to the right."}
	endif
	ModifyControlList ControlNameList(pName,";","bLayer_*") fsize=14 ,fstyle=1
	
	if (!CmpStr(IgorInfo(2), "Macintosh"))
		ModifyControlList ControlNameList(pName,";","*") fsize=10
	else
		ModifyControlList ControlNameList(pName,";","bLayer_Rot*") font="Segoe UI Symbol"
	endif
	
	MakeButtonIntoWSPopupButton(pName, "LoadProfile", "SpectraImageProfile#LoadProfileAction", popupWidth=230, popupHeight=400, content = WMWS_Waves, options = PopupWS_OptionTitleInTitle)
	PopupWS_MatchOptions(pName, "LoadProfile", listoptions="DIMS:1,TEXT:0,WAVE:0,DF:0,CMPLX:0")
	PopupWS_AddSelectableString( pName, "LoadProfile", "none")
	PopupWS_SetSelectionFullPath(pName, "LoadProfile", "none")
	
	ModifyControlList ControlNameList(pName, ";", "b*")  proc=Profile_MainButtonProc
	Button bNormUser			,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)			// activate measurement info normalization if the conditions are right
	Button bSaveSpec			,fColor=(65535,0,0)
	CheckBox IntColCheckBox		,value = 1 ,mode = 1	,proc=ModeCheckBoxControl				// radio button checked => horizontal mode
	CheckBox IntRowCheckBox		,value = 0 ,mode = 1	,proc=ModeCheckBoxControl				// radio button unchecked
	CheckBox FullWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MinWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox Normto1			,value = 0				,proc=ModeCheckBoxControl
	CheckBox SnapToGrid			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MousetoSpec		,value = 0				,proc=ModeCheckBoxControl
	CheckBox MaintainInt		,value = 0
	CheckBox cAlignSmooth		,value = 0
	CheckBox cAlignOutput		,value = 0
	CheckBox cAlignSubShift		,value = 1
	CheckBox cAlignFillGaps		,value = 0
	CheckBox cDoLogScale		,value = 0 ,side = 1	,proc=Profile_DisplayCheckBox
	CheckBox cPrintEdits		,value = kEnableLogPrintout
	SetVariable SubtractScale	,bodywidth = 40			,limits={-inf,inf,0.1}	,value=_NUM:1
	SetVariable NormSmooth		,bodywidth = 40			,limits={0,32767,1}		,value=_NUM:2
	SetVariable Position		,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable Width			,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable SmoothFactor	,bodywidth = 65			,limits={0,32767,1}		,value=_NUM:(ceil(DimSize(inwave,0)*kAlignSmooth/100))
	PopupMenu ContrastColor		,bodyWidth = 160		,mode=1+WhichListItem("Terrain", CTabList())	,proc=Profile_ContrastColorProc
	
	Duplicate/Free/R=[][][0] inwave, tempLayer;	WaveStats/Q tempLayer
	Slider ContrastLow			,value=V_min			,limits={V_min,V_max,0}	,proc=Profile_ContrastSliderProc
	Slider ContrastHigh			,value=V_max			,limits={V_min,V_max,0}	,proc=Profile_ContrastSliderProc
	//+++++++++++++++++++++++++++++++++++++ set dependency +++++++++++++++++++++++++++++++++++++++
	Profile_AppendSectionTrace(pName)															// add the ROI lines & set variable limits for the first time
	Profile_UpdateAxes(pName, -2)																// flips the axes according to the wave's delta scaling (-2 means 'reset')
	Profile_SetMinFullLineValues(pName,1,0,0)													// set initial values for row mode (position & width)
	Profile_SetMinFullLineValues(pName,1,1,0)													// set initial values for col mode

	Profile_UpdateLinesAndProfile(pName)														// first update to generate initial profile
	Profile_SetFullMinCheckbox(pName)
	DoWindow/F $gName
	SetActiveSubwindow ##																		// Make sure the graph can be acted upon
	return 0
end

//################################################################################################

static function/DF getWorkFolder(string pName)
	string folder = ""
	if (strlen(pName))
		folder = GetUserData(pName,"","workFolder")
	endif
	return root:Packages:$(kWorkingDir):$(folder)
end

static function/S getPanel(string gName)
	string subPanel = ListMatch(ChildWindowList(gName), "*Profile*")
	if(strlen(subPanel))
		return gName+"#"+StringFromList(0,subPanel)
	elseif (WinType("ProfileGeneratorWindow"))													// compatibility with older versions
		return "ProfileGeneratorWindow"
	else
		return ""
	endif
end

static function/S getGraph(string pName)
	return StringFromList(0,pName,"#")
end

//################################################################################################

static function LoadProfileAction(variable event, string wavePath, string pName, string ctrlName)
	DFREF stg = getWorkFolder(pName)
	wave para = stg:CurrentLines
	string load = "Load"+SelectString(para[0],"Horiz","Vert")+"WaveLoc"
	if (!CmpStr(wavePath,"none"))
		KillStrings/Z stg:$load
	elseif (Exists(wavePath) == 1)
		string/G stg:$load = wavePath
	else
		return 0
	endif
	Profile_UpdateLinesAndProfile(pName)
	Profile_AppendSectionTrace(pName)
end

//################################################################################################

static function Profile_QuitProgram(string inPath)
	string gName = getGraph(inPath)																// can be called from AskQuitPanel as well => construct panel path manually
	string pName = getPanel(gName)
	DFREF stg = getWorkFolder(pName)
	DFREF main= getWorkFolder("")
	DFREF pkg = root:Packages:
	
	SetWindow $pName hook(CreateProfile)=$""													// release hook to not trigger any events
	SetWindow $gName hook(CreateProfile)=$""
	KillWindow $getGraph(pName)																	// first, close image window
	NVar/Z ProfileUpdateConnector = stg:ProfileUpdateConnector
	if (!NVAR_Exists(ProfileUpdateConnector))
		return -1																				// obviously the folder does not exist anymore
	endif
	SetFormula ProfileUpdateConnector,""														// reset formula
	KillDataFolder/Z root:WinGlobals:$(gName)
	if (CountObjectsDFR(root:WinGlobals, 4) == 0)
		KillDataFolder/Z root:WinGlobals
	endif
	if (CountObjectsDFR(stg, 4) == 0)
		KillDataFolder/Z stg
	endif
	variable Objects = CountObjectsDFR(main, 1)+CountObjectsDFR(main, 2)+CountObjectsDFR(main, 3)+CountObjectsDFR(main, 4)
	if (Objects == 0 && DataFolderRefStatus(main) != 0)
		KillDataFolder main
	endif
	Objects = CountObjectsDFR(pkg, 1)+CountObjectsDFR(pkg, 2)+CountObjectsDFR(pkg, 3)+CountObjectsDFR(pkg, 4)
	if (Objects == 0)																			// clean up Packages folder if empty
		KillDataFolder pkg
	endif
	return 0
end

//################################################################################################

static function Profile_AppendSectionTrace(string pName)										// appends the profile parallel to the selection lines
	string gName = getGraph(pName)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave   para = stg:CurrentLines
	wave   SecL = stg:SectionLine
	wave/Z cmpL	= stg:CompareLine
	wave/Z multi= stg:MultiSection
	int i, mode = para[0]
	wave/Z load	= stg:$("LoadProfile_"+SelectString(mode,"Horiz","Vert"))
	
	Profile_SetPosWidthControls(pName)															// initialize center and width controls
	
	string getSplit																				// axes may have been dragged => get new split location
	getSplit = GetUserData(pName, "", "vertAxSplit");	variable vertAx = strlen(getSplit) ? str2num(getSplit) : 0.3
	getSplit = GetUserData(pName, "", "horiAxSplit");	variable horiAx = strlen(getSplit) ? str2num(getSplit) : 0.5

	//+++++++++++++++++++++++++++++++++ multi-profile preparation ++++++++++++++++++++++++++++++++
	string list = GrepList(TraceNameList(gName,";",1), "^(?!Lines)", 0) 						// first, remove all
	for (i = ItemsInList(list)-1; i > -1; i--)
		RemoveFromGraph/W=$gName/Z $StringFromList(i,list)	
	endfor
	
	ControlInfo/W=$pName MultiProfile;	int doMulti = V_flag==2 && V_disable==0 && V_Value==1 && WaveExists(multi)
	if (doMulti)
		variable row, size = min(DimSize(multi,1)-1,DimSize($wLoc,2)-1)							// 'delete layer' may have reduced the size
		ColorTab2Wave $kMultiProfileColor;	wave M_colors
		for (i = 0; i < size+1; i++)										
			if (mode)	// vertical
				AppendToGraph/W=$gName/VERT/B=bottom2 multi[][size-i]							// go backwards (in reverse order, so that the later traces are BEHIND the earlier ones)
			else		// horizontal
				AppendToGraph/W=$gName/L=left2 multi[][size-i]
			endif
			row = (i/size)*(DimSize(M_Colors, 0)-1)
			ModifyGraph/W=$gName rgb($("MultiSection"+SelectString(i>0,"","#"+num2str(i))))=(M_Colors[row][0], M_Colors[row][1], M_Colors[row][2])
		endfor
		KillWaves/Z M_colors
	endif
	
	//+++++++++++++++++++++++++++++++++++++++ main profile +++++++++++++++++++++++++++++++++++++++
	ControlInfo/W=$pName cDoLogScale;	int doLog = V_flag==2 ? V_Value : 0
	if (mode)	// vertical
		AppendToGraph/W=$gName/VERT/B=bottom2 SecL
		ModifyGraph/W=$gName axisEnab(left)={0,1},	axisEnab(bottom) ={vertAx+0.01,1}			// modify the main axes
		ModifyGraph/W=$gName freePos(bottom2)=0,	axisEnab(bottom2)={0,vertAx-0.01}			// align the new bottom axis
		ModifyGraph/W=$gName zero(bottom2)=4, lblPosMode(bottom2)=1, log(bottom2)=doLog
		if (WaveExists(cmpL))
			AppendToGraph/W=$gName/VERT/B=bottom2 cmpL
			ModifyGraph/W=$gName rgb(CompareLine)=(0,35000,0), lsize(CompareLine)=(1 + doMulti)
		endif
		if (WaveExists(load))
			AppendToGraph/W=$gName/VERT/B=bottom2 load
			ModifyGraph/W=$gName rgb(LoadProfile_Vert)=(34000,10000,60000), lsize(LoadProfile_Vert)=(1 + doMulti)
		endif
		ColorScale/W=$gName/C/N=ImageColorDisplay/F=0/Z=1/A=LB/X=((vertAx+0.01)*100)/Y=100.10 " "		// debug: add /B=(65535,0,0)
		ColorScale/W=$gName/C/N=ImageColorDisplay image=$StringFromList(0,ImageNameList(gName, ";"))
		ColorScale/W=$gName/C/N=ImageColorDisplay vert=0, side=2, widthPct=100-(vertAx+0.01)*100, height=8, tickLen=5
		ColorScale/W=$gName/C/N=ImageColorDisplay tickLblRot=0, fsize=10, lblMargin=0, minor=1, tickUnit=1, log=doLog
		
		ModifyGraph/W=$gName mirror=1
		ModifyGraph/W=$gName margin(top)=40, margin(right)=0, mirror(bottom)=2
	else		// horizontal
		AppendToGraph/W=$gName/L=left2 SecL
		ModifyGraph/W=$gName axisEnab(bottom)={0,1}, axisEnab(left) ={horiAx+0.01,1}
		ModifyGraph/W=$gName freePos(left2)=0,		 axisEnab(left2)={0,horiAx-0.01}
		ModifyGraph/W=$gName zero(left2)=4, lblPosMode(left2)=1, log(left2)=doLog
		if (WaveExists(cmpL))
			AppendToGraph/W=$gName/L=left2 cmpL
			ModifyGraph/W=$gName rgb(CompareLine)=(0,35000,0), lsize(CompareLine)=(1 + doMulti)
		endif
		if (WaveExists(load))
			AppendToGraph/W=$gName/L=left2 load
			ModifyGraph/W=$gName rgb(LoadProfile_Horiz)=(34000,10000,60000), lsize(LoadProfile_Horiz)=(1 + doMulti)
		endif
		ColorScale/W=$gName/C/N=ImageColorDisplay/F=0/Z=1/A=LT/X=100.10/Y=0.00 " "						// debug: add /B=(65535,0,0)
		ColorScale/W=$gName/C/N=ImageColorDisplay image=$StringFromList(0,ImageNameList(gName, ";"))
		ColorScale/W=$gName/C/N=ImageColorDisplay vert=1, side=1, heightPct=100-(horiAx+0.01)*100, width=8, tickLen=5
		ColorScale/W=$gName/C/N=ImageColorDisplay tickLblRot=90, fsize=10, lblMargin=0, minor=1, tickUnit=1, log=doLog
		
		ModifyGraph/W=$gName mirror=1
		ModifyGraph/W=$gName margin(top)=0, margin(right)=40, mirror(left)=2
	endif
	ModifyGraph/W=$gName highTrip=100000, standoff=0, minor=1
	ModifyGraph/W=$gName ZisZ=1, notation=1, rgb(SectionLine)=(65280,0,0)
	ModifyGraph/W=$gName lsize(SectionLine)=(1 + doMulti)
	
	if (doMulti)																				// also add color indicator for multi-profiles
		if (mode)
			ModifyGraph mirror(bottom2)=2
			ColorScale/W=$gName/C/N=ProfileColorDisplay/F=0/Z=1/A=LB/X=0.00/Y=100.10
			ColorScale/W=$gName/C/N=ProfileColorDisplay ctab={0,size,$kMultiProfileColor,1} ,nticks=limit(size,1,100)
			ColorScale/W=$gName/C/N=ProfileColorDisplay vert=0, side=2, widthPct=(vertAx-0.01)*100, height=8, tickLen=5, fsize=10, lblMargin=0
		else
			ModifyGraph mirror(left2)=2
			ColorScale/W=$gName/C/N=ProfileColorDisplay/F=0/Z=1/A=LB/X=100.10/Y=0.00
			ColorScale/W=$gName/C/N=ProfileColorDisplay ctab={0,size,$kMultiProfileColor,1} ,nticks=limit(size,1,100)
			ColorScale/W=$gName/C/N=ProfileColorDisplay vert=1, side=1, heightPct=(horiAx-0.01)*100, width=8, tickLen=5, fsize=10, lblMargin=0
		endif
	else
		ColorScale/K/N=ProfileColorDisplay
	endif
	
	return 0
end

//################################################################################################

static function Profile_ShiftMultiProfile(string pName)											// shifts all profiles to to their appropriate places
	string gName = getGraph(pName)
	string sPath = gname+"#floatPanel#zProfile"
	DFREF stg = getWorkFolder(pName)
	wave/Z multi = stg:MultiSection
	
	ControlInfo/W=$pName SetLayer;		variable layer  = V_Value
	ControlInfo/W=$pName vMulti_Dist;	variable offset = V_Value/100
	
	if (WinType(sPath) && WhichListItem("SectionLine_Z",TraceNameList(sPath,";",1)) > -1)		// layer indicator in the z-profile graph
		Cursor/W=$sPath/P D, SectionLine_Z, layer
	endif
	
	ControlInfo/W=$pName MultiProfile															// check for the existence of these controls
	if (V_flag != 2 || V_disable != 0 || V_Value != 1 || !WaveExists(multi))
		return 0
	endif
	
	WaveStats/Q multi
	offset *= sign(V_min)!=sign(V_max) ? abs(V_min-V_max) : max(abs(V_max),abs(V_min))			// trace offsets are in percent of the total range or the maximum 
	string tlist = GrepList(TraceNameList(gName,";",1), "^MultiSection", 0)
	int i, items = ItemsInList(tlist)
	for (i = 0; i < items; i++)
		ModifyGraph/W=$gName offset($StringFromList(items-1 - i,tlist))={,i*offset}				// offset all multi-profiles
		ModifyGraph/W=$gName  lsize($StringFromList(items-1 - i,tlist))=!(i==layer)
	endfor
	
	tlist = GrepList(TraceNameList(gName,";",1), "^(?!Multi|Lines)", 0)							// All other normal traces
	items = ItemsInList(tlist)
	for (i = 0; i < items; i++)
		ModifyGraph/W=$gName offset($StringFromList(i,tlist))={,layer*offset}
	endfor
	return 0
end

//################################################################################################

static function Profile_SetPosWidthControls(string pName)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc
	int  mode = para[0]
	
	string unit = WaveUnits(data,1-mode);	unit[0] = SelectString(strlen(unit),""," ")
	variable delta = abs(DimDelta(data,1-mode))
	variable init = delta*round((DimSize(data,1-mode)-1) / 25)									// calculate ca. 5% of width
	init = init <= delta ? 2*delta : init														// make sure that's not too small

	SetVariable Position ,win=$pName ,limits={-inf,inf,init}	,value=para[1+mode*2]	,format="%g"+unit	// set new control values	
	ControlInfo/W=$pName SnapToGrid;   init = V_Value ? init+mod(init/delta,2)*delta : init					// make sure the change is even when snapped to the grid (=> symmetric width change)
	SetVariable Width	 ,win=$pName ,limits={delta,inf,init}	,value=para[2+mode*2]	,format="%g"+unit
	return 0
end

//################################################################################################

function Profile_ProcessWindowEvent(STRUCT WMWinHookStruct &s)
	string gName = getGraph(s.winName)															// can be called from AskQuitPanel as well => construct panel path manually
	string pName = getPanel(gName)
	DFREF stg = getWorkFolder(pName)
	if (DataFolderRefStatus(stg) == 0)															// something is wrong if the data is not found => exit
		return -1
	endif
	SVAR/Z wLoc = stg:InwaveLocation
	wave/Z para = stg:CurrentLines
	if (!SVAR_Exists(wLoc))																		// something is wrong => maybe the graph was exported
		KillWindow/Z $pName
		return 0
	endif
	wave data = $wLoc
	
	int hookReturn = 0, mode = para[0]
	if (!CmpStr(s.eventName,"keyboard"))
		//+++++++++++++++++++++++++++++++++ execute button keys ++++++++++++++++++++++++++++++++++
		string Key = ""
		Switch(s.keyCode)
			case 112:	// p key
				Key = "bCompare"
			break
			case 115:	// s key
				Key = "bSaveSpec"
			break
			case 117:	// u key
				Key = "bUndoAll"
			break
			case 127:	// del key
				Key = "bRemoveSelection"
			break
		EndSwitch
		if (strlen(Key) > 0)
			Profile_ExecuteButtonCommand(pName, Key)
			return 1
		endif
		//++++++++++++++++++++++++++++++++ execute checkbox keys ++++++++++++++++++++++++++++++++++
		int checkIt = 1
		Switch(s.keyCode)
			case 11:	// page up
				Key = "FullWidth"
			break
			case 12:	// page down
				Key = "MinWidth"
			break
			case 103:	// g key
				Key = "toggleSnapToGrid"
			break
			case 104:	// h key
				Key = "IntColCheckBox"
				CheckBox $Key ,value=checkIt ,win=$pName
				DrawAction/W=$gName getgroup=axisSplitLine, delete
			break
			case 109:	// m key
				Key = "toggleMousetoSpec"
			break
			case 110:	// n key
				Key = "toggleNormto1"
			break
			case 118:	// v key
				Key = "IntRowCheckBox"
				CheckBox $Key ,value=checkIt ,win=$pName
				DrawAction/W=$gName getgroup=axisSplitLine, delete
			break
			case 122:	// z key
				ControlInfo/W=$pName doZProfiles
				if (V_flag == 2)
					checkIt = !v_value
					Key = "doZProfiles"
					CheckBox $Key ,value=checkIt ,win=$pName
				endif
			break
		EndSwitch
		if (strlen(Key) > 0)
			Profile_ExecuteCheckboxAction(pName, Key, checkIt)
			return 1
		endif
		//+++++++++++++++++++++++++++++++++ execute arrow keys +++++++++++++++++++++++++++++++++++
		ControlInfo/W=$pName SnapToGrid
		variable Shift = 0, Expand = 0, snapLines = V_Value
		Switch (mode)																			// define a shift and expand direction (positive or negative) depending on mode and pressed key
			case 0:
				Shift	= s.keyCode == 30 ? 1 : ( s.keyCode == 31 ? -1 : Shift )
				Expand	= s.keyCode == 29 ? 1 : ( s.keyCode == 28 ? -1 : Expand )
			break
			case 1:
				Shift	= s.keyCode == 29 ? 1 : ( s.keyCode == 28 ? -1 : Shift )
				Expand	= s.keyCode == 30 ? 1 : ( s.keyCode == 31 ? -1 : Expand )
			break
		EndSwitch
		
		if (Shift != 0 || Expand != 0)
			hookReturn = 1
			variable delta = DimDelta(data,1-mode)
			variable boost = ceil(DimSize(data,1-mode)/100)										// move further for bigger waves
			variable speed = 1																	// add acceleration modifier depending on the pressed modifier key
			speed = (s.eventMod & 2^1) != 0 ? 4 : speed
			speed = (s.eventMod & 2^3) != 0 ? 0.5 : speed
			speed = ((s.eventMod & 2^3) && (s.eventMod & 2^1)) != 0 ? 1 : speed					// move exactly one delta unit with both shift and ctrl pressed
			boost = ((s.eventMod & 2^3) && (s.eventMod & 2^1)) != 0 ? 1 : boost
			
			para[1 + mode*2] += Shift * delta * speed * boost									// do the shift
			para[2 + mode*2] += Expand* (1+snapLines) * abs(delta) * speed * boost				// do the expand (even expansion for snapped mode)
			para[2 + mode*2]  = para[2 + mode*2] < abs(delta) ? abs(delta) : para[2 + mode*2]	// check for the minimal width (will normally be done by the variable control)
			
			Profile_CheckBordersAndUpdateAll(pName)												// check if the lines go over the borders of the wave
			Profile_SetFullMinCheckbox(pName)													// toggle check-boxes
		endif
	endif
	
	if (WinType(s.WinName) == 7)																// process only key presses for panels! otherwise mouse and kill events are passed through
		return hookReturn
	endif
	
	//############################################################################################

	string sName = gname+"#floatPanel#zProfile", saveStruct = ""
	GetWindow $gName, expand
	variable gExpand = V_value;	gExpand = gExpand == 0 ? 1 : gExpand							// pixel conversion and expansion factor
	variable pixConv = ScreenResolution/72*gExpand
	
	if (WhichListItem(s.eventName, "mousedown;mousemoved;mouseup;")>-1)							// mouse events on the graph
		variable axDiv = 0, dragPoint = 0
		variable axDist = 5, edgeDist = 15														// fixed proximity ranges
		if (!CmpStr(s.winName, gName))
			sscanf StringByKey("axisEnab(x)",AxisInfo(gName,SelectString(mode,"left","bottom")),"="), "{%f,", axDiv				// dynamic divider line
			GetWindow $gName, psizeDC;	axDiv -= 0.01																			// center split position
			variable mouseDelta = mode ? (s.mouseloc.h-v_left) : (s.mouseloc.v-v_top)
			variable plotRange  = mode ? (v_right     -v_left) : (v_bottom    -v_top)
			variable mouseClose = mode ? abs(mouseDelta-plotRange*axDiv)<axDist : abs(mouseDelta-plotRange*(1-axDiv))<axDist
			variable relatMouse = mode ?     mouseDelta/plotRange               : 1 - mouseDelta/plotRange						// relative mouse position on the plot
		endif
		if (!CmpStr(s.winName, sName))
			if (!CheckMouseInPlot(s.mouseLoc,sName))											// mouse is not in plot area
				GetWindow $sName gsizeDC;	 dragPoint = 1										// encode sub-window mouse edge detection into dragPoint
				dragPoint = abs(s.mouseloc.h-v_left) <edgeDist && abs(s.mouseloc.v-v_top)   <edgeDist ? 2 : dragPoint	// LT
				dragPoint = abs(s.mouseloc.h-v_right)<edgeDist && abs(s.mouseloc.v-v_top)   <edgeDist ? 3 : dragPoint	// RT
				dragPoint = abs(s.mouseloc.h-v_left) <edgeDist && abs(s.mouseloc.v-v_bottom)<edgeDist ? 4 : dragPoint	// LB
				dragPoint = abs(s.mouseloc.h-v_right)<edgeDist && abs(s.mouseloc.v-v_bottom)<edgeDist ? 5 : dragPoint	// RB
			endif
		endif
	endif
	
	STRUCT DragStruct d																			// sub-window drag structures
	STRUCT rect tempRect
	
	StrSwitch (s.eventName)
		case "killVote":																		// window close button
			hookReturn = 1
			if (EqualWaves(stg:startCopy, data,-1))												// compare if the wave was changed
				Profile_ExecuteButtonCommand(pName, "bQuit")									// just quit (no undo)
				break
			endif
			if (Wintype("#AskQuitPanel"))
				break
			endif
			
			sName = gName+"#AskQuitPanel"														// this will create a little panel in case of changes
			variable sizePix = (!CmpStr(IgorInfo(2), "Windows") && ScreenResolution == 96)? 72/ScreenResolution : 1
			variable pW_half = 112*sizePix/gExpand, pH_half = 35*sizePix/gExpand
			variable xCenter = (s.WinRect.right - s.WinRect.left)/2/pixConv
			variable yCenter = (s.WinRect.bottom - s.WinRect.top)/2/pixConv
			
			NewPanel/N=AskQuitPanel/HOST=$gName/W=(xCenter-pW_half,yCenter-pH_half,xCenter+pW_half,yCenter+pH_half)
			DrawText/W=$sName 40,28,"Accept Changes to Data?"
			Button bAskOK		,win=$sName	,pos={ 12,33} ,size={60,23} ,title="Yes"
			Button bAskUndo		,win=$sName	,pos={ 82,33} ,size={60,23} ,title="Undo"
			Button bAskCancel	,win=$sName	,pos={152,33} ,size={60,23} ,title="Cancel"
				ModifyControlList ControlNameList(sName,";","*") proc=Profile_AskButton
			if (CmpStr(IgorInfo(2), "Macintosh") == 0)
				ModifyControlList ControlNameList(sName,";","*") fsize=10
			endif
			SetActiveSubwindow $gName															// restore focus to main window
		break
		case "mousedown":
			if (s.eventMod !=1)																	// override graph interaction only when normal mouse button is pressed
				break
			endif
			//+++++++++++++++++++++++++++++++++ axis quick drag ++++++++++++++++++++++++++++++++++
			if (mouseClose)																		// initialize axis drag state
				SetWindow $pName userdata(dragAxis)="1"
				hookReturn = 1
			endif
			//++++++++++++++++++++++++++++++ drag floating panel ++++++++++++++++++++++++++++++++
			if (CmpStr(s.winName, sName) || CheckMouseInPlot(s.mouseLoc,sName))
				break
			endif
			d.mouseVec = s.mouseloc
			d.subwRect = s.winrect
			StructPut/S d, saveStruct
			SetWindow $pName userdata(dragStruct)=saveStruct
			SetWindow $pName userdata(dragFloat)=num2str(dragPoint)
			hookReturn = 1
		break
		case "mouseup":
			if (str2num(GetUserData(pName,"","dragFloat")))										// bring main graph to front after dragging sub-window
				SetActiveSubwindow $gName
			endif
			SetWindow $pName userdata(dragAxis) ="0"											// release axis drag
			SetWindow $pName userdata(dragFloat)="0"
		break
		case "mousemoved":
			variable dragEdge = str2num(GetUserData(pName, "", "dragFloat"));	dragEdge = numtype(dragEdge) != 0 ? 0 : dragEdge
			variable dragAxis = str2num(GetUserData(pName, "", "dragAxis"));	dragAxis = numtype(dragAxis) != 0 ? 0 : dragAxis
			if (dragAxis || dragEdge)
                GetMouse/W=$gName
                if (!(V_flag & 1))																// mouse button released? => maybe we missed a mouse-up event?
                    SetWindow $pName userdata(dragAxis) ="0";	dragAxis = 0
                    SetWindow $pName userdata(dragFloat)="0";	dragEdge = 0
                endif
            endif
			
			//+++++++++++++++++++++++++++++++++ axis quick drag ++++++++++++++++++++++++++++++++++
			DrawAction/W=$gName getgroup=axisSplitLine, delete									// draws and sets divider line for dynamic axis drag
			if (dragAxis)
				axDiv = limit(relatMouse, 0.15, 0.85)											// set new divider
				ModifyGraph/W=$gName axisEnab($SelectString(mode,"left","bottom"))	 = {axDiv + 0.01, 1}
				ModifyGraph/W=$gName axisEnab($SelectString(mode,"left2","bottom2")) = {0, axDiv - 0.01}
				SetWindow $pName userdata($SelectString(mode,"horiAxSplit","vertAxSplit"))=num2str(axDiv)	// save current position into panel
				if (FindListItem("ImageColorDisplay", AnnotationList(gName)) > -1)							// scale image color legend
					if (mode)
						ColorScale/W=$gName/C/N=ImageColorDisplay/X=((axDiv+0.01)*100) widthPct=100-(axDiv+0.01)*100
					else
						ColorScale/W=$gName/C/N=ImageColorDisplay heightPct=100-(axDiv+0.01)*100
					endif
				endif
				if (FindListItem("ProfileColorDisplay", AnnotationList(gName)) > -1)			// scale profile color legend as well
					if (mode)
						ColorScale/W=$gName/C/N=ProfileColorDisplay  widthPct=(axDiv-0.01)*100
					else
						ColorScale/W=$gName/C/N=ProfileColorDisplay heightPct=(axDiv-0.01)*100
					endif
				endif
				hookReturn = 1
			endif
			
			if (dragAxis || mouseClose)															// draw a line for visualization
				SetDrawEnv/W=$gName linefgc=(36000,15000,59000) ,dash=2 ,linethick= 1.50
				SetDrawEnv/W=$gName gstart, gname=axisSplitLine
				if (mode)
					DrawLine/W=$gName axDiv, -0.1, axDiv, 1.1
				else
					DrawLine/W=$gName -0.1, 1-axDiv, 1.1, 1-axDiv
				endif
				SetDrawEnv/W=$gName gstop, gname=axisSplitLine
				s.cursorCode = mode ? 5 : 6														// change mouse cursor
				s.doSetCursor = 1
			endif
			
			//+++++++++++++++++++++++++++++++ drag floating panel ++++++++++++++++++++++++++++++++
			if (dragPoint)
				s.cursorCode = str2num(StringFromList(dragPoint,"0;8;4;11;10;7;"))				// change mouse cursor
				s.doSetCursor = 1
			endif
			
			if (dragEdge)
				StructGet/S d, GetUserData(pName, "", "dragStruct")
				variable deltah = s.mouseLoc.h - d.mouseVec.h
				variable deltav = s.mouseLoc.v - d.mouseVec.v
				tempRect = d.subwRect
				tempRect.top    += deltav * (dragEdge == 1 || dragEdge == 2 || dragEdge == 3)	// decide which edge to drag
				tempRect.left   += deltah * (dragEdge == 1 || dragEdge == 2 || dragEdge == 4)
				tempRect.right  += deltah * (dragEdge == 1 || dragEdge == 5 || dragEdge == 3)
				tempRect.bottom += deltav * (dragEdge == 1 || dragEdge == 5 || dragEdge == 4)
				if (CheckRectPosAndSize(tempRect, gName))
					MoveSubwindow/W=$(gName+"#floatPanel") fnum=(tempRect.left/pixConv, tempRect.top/pixConv, tempRect.right/pixConv, tempRect.bottom/pixConv)
					d.subwRect = tempRect
				endif
				d.mouseVec = s.mouseLoc
				StructPut/S d, saveStruct
				SetWindow $pName userdata(dragStruct)=saveStruct
			endif
			
			//++++++++++++++++++++++++++++++ quick mouse profile ++++++++++++++++++++++++++++++++
			ControlInfo/W=$pName MousetoSpec
			if (!V_Value)
				break
			endif
			
			ControlInfo/W=$pName SetLayer;	int layer = abs(V_flag) == 5 ? V_Value : 0			// get the current layer
			ControlInfo/W=$pName Normto1;	int doNorm = V_Value								// also normalize the mouse profile
			variable wmin = 0, wmax = 1
			variable Mx = round ((AxisValFromPixel("", "Bottom", s.mouseLoc.h)-DimOffset(data,0)) / DimDelta(data,0))
			variable My = round ((AxisValFromPixel("", "Left",   s.mouseLoc.v)-DimOffset(data,1)) / DimDelta(data,1))
			if (Mx >= DimSize(data, 0) || My >= DimSize(data, 1) || Mx < 0  || My < 0)			// mouse outside of wave
				break
			endif
			
			RemoveFromGraph/W=$gName/Z MouseProfile
			if (mode)																			// append one row / column under the current mouse position
				AppendToGraph/W=$gName/VERT/B=bottom2 data[Mx][][layer]/TN=MouseProfile
			else
				AppendToGraph/W=$gName/L=left2 data[][My][layer]/TN=MouseProfile
			endif
			if (doNorm)
				if (mode)
					Duplicate/Free/R=[Mx][][layer] data, tempProfile
				else
					Duplicate/Free/R=[][My][layer] data, tempProfile
				endif
				WaveStats/Q tempProfile; wmin = V_min; wmax = V_max
			endif
			
			if (WinType(sName))
				RemoveFromGraph/W=$sName/Z MouseProfile
				AppendToGraph/W=$sName data[Mx][My][]/TN=MouseProfile
				ModifyGraph/W=$sName muloffset(MouseProfile)={,1/(wmax-wmin)}
				ModifyGraph/W=$sName       rgb(MouseProfile)=(0,39321,0)
			endif
			
			sscanf StringByKey("offset(x)",TraceInfo(gName,"SectionLine",0),"="), "{%f,%f}", Mx, My		// bring the mouse profile to the same height as the main profile
			ModifyGraph/W=$gName    offset(MouseProfile)={,My-wmin/(wmax-wmin)}
			ModifyGraph/W=$gName muloffset(MouseProfile)={,1/(wmax-wmin)}
			ModifyGraph/W=$gName       rgb(MouseProfile)=(0,39321,0)
		break
		case "resize":
			if (WinType(sName) || !CmpStr(s.winName,sName))										// make sure the sub-window stays within range
				GetWindow $gName wsizeDC
				int isHostMinW = abs(v_right-v_left) <= subPanelMin
				int isHostMinH = abs(v_bottom-v_top) <= subPanelMin
				GetWindow $sName wsizeDC
				variable subW = abs(v_right-v_left)
				variable subH = abs(v_bottom-v_top)
				variable dW = isHostMinW ? 0 : subPanelMin - subW								// check if the panel width is smaller than the minimum
				variable dH = isHostMinH ? 0 : subPanelMin - subH
				if (dW <= 0 && dH <= 0)
					break
				endif
				
				string dragInfo = GetUserData(pName,"","dragStruct")
				if (strlen(dragInfo))
					StructGet/S d, dragInfo
					subW = abs(d.subwRect.right -d.subwRect.left)
					subH = abs(d.subwRect.bottom-d.subwRect.top)
				endif
				tempRect.top  = v_top - dH*(dH>0); tempRect.bottom = tempRect.top + subH
				tempRect.left = v_left- dW*(dW>0); tempRect.right  = tempRect.left+ subW
				MoveSubwindow/W=$(gname+"#floatPanel") fnum=(tempRect.left/pixConv, tempRect.top/pixConv, tempRect.right/pixConv, tempRect.bottom/pixConv)
			endif
		break
	EndSwitch
	return hookReturn
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function Profile_AskButton(STRUCT WMButtonAction &s) : ButtonControl							// execute button commands of the quit sub-panel
	if (s.eventCode == 2)
		Strswitch (s.ctrlName)
			case "bAskUndo":
				Profile_ExecuteButtonCommand(getPanel(getGraph(s.win)), "bUndoAll")
			case "bAskOK":
				Profile_ExecuteButtonCommand(getPanel(getGraph(s.win)), "bSaveQuit")
			break
			case "bAskCancel":																	// just remove the panel
				KillWindow #AskQuitPanel
			break
		EndSwitch
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Constant subPanelMin = 100
static structure DragStruct
	STRUCT point mouseVec
	STRUCT Rect subwRect
endstructure

static function CheckRectPosAndSize(STRUCT rect &in, string host)
	GetWindow $host wsizeDC
	int isInside = (in.left>v_left && in.top>v_top && in.right<v_right && in.bottom<v_bottom)
	int isMinDim = abs(in.right-in.left) >= subPanelMin && abs(in.bottom-in.top) >= subPanelMin
	return isInside && isMinDim
end

static function CheckMouseInPlot(STRUCT point &pnt, string host)
	GetWindow $host psizeDC
	return (pnt.h>v_left-1 && pnt.h<v_right+1 && pnt.v>v_top-1 && pnt.v<v_bottom+1)				// a bit smaller than the plot area
end

//################################################################################################

static function Profile_UpdateLinesAndProfile(string pName)										// generates marquee lines
	DFREF stg = getWorkFolder(pName)
	wave LinesX	= stg:LinesX
	wave LinesY	= stg:LinesY
	wave para = stg:CurrentLines
	variable plus, minus
	if (para[0])																				// depending on mode switch line direction
		plus	= para[3] + para[4]/2
		minus	= para[3] - para[4]/2
		LinesX	= {plus,plus,NaN,minus,minus}
		LinesY	= {-INF,INF,NaN,-INF,INF}
	else
		plus	= para[1] + para[2]/2
		minus	= para[1] - para[2]/2
		LinesX	= {-INF,INF,NaN,-INF,INF}
		LinesY	= {plus,plus,NaN,minus,minus}
	endif
	ModifyGraph/W=$getGraph(pName) offset(LinesY)={0,0}											// reset offset, trigger S_TraceOffset
	return 0
end

//################################################################################################

static StrConstant kZProfileModes = "Average;Maxima;Minima;"

function LinesIntegrateUpdate(string Info)														// will be triggered when S_TraceOffset changes (connected via formula)
	if( StrSearch(Info,"TNAME:LinesY;",0) <=0 )													// contain the Info line data?
		return 0
	endif
	string gName = StringByKey("GRAPH",Info)
	string pName = getPanel(gName)
	DFREF stg = getWorkFolder(pName)
	if (!DataFolderRefStatus(stg))																// something is wrong if the data is not found => exit
		return -1
	endif
	SVAR wLoc = stg:InwaveLocation
	if (!SVAR_Exists(wLoc))
		return -1
	endif
	
	wave   para = stg:CurrentLines
	wave   secL = stg:SectionLine
	wave/Z multi= stg:MultiSection
	wave   data = $wLoc
	int	i, mode = para[0]
	if (mode == 0)																				// depending on mode (in row [0]) read the line data (X or Y direction)
		para[1] += NumberByKey("YOFFSET",Info)
	else
		para[3] += NumberByKey("XOFFSET",Info)
	endif
	
	Profile_CheckBordersAndUpdateAll(pName)														// check if the lines go over the borders of the wave before calculating something
	//++++++++++++++++++++++++++++++++++++ prepare integration ++++++++++++++++++++++++++++++++++
	ControlInfo/W=$pName SetLayer
	variable layer	= abs(V_flag) == 5 ? V_Value : 0
	variable center	= para[1 + mode * 2]
	variable width	= para[2 + mode * 2]
	variable delta	= DimDelta(data,  1-mode)
	variable offset	= DimOffset(data, 1-mode)
	variable size	= DimSize(data,   1-mode)
	string unit		= WaveUnits(data, 1-mode);	unit[0] = SelectString(strlen(unit),""," ")

	string NoteText = note(data) + "\r" + NameOfWave(data) + " integrated"
	NoteText += " from "+num2str(ChangePrecision(center-width/2,4))
	NoteText += " to "	+num2str(ChangePrecision(center+width/2,4))+" "+unit+" (width = "+num2str(width)+" "+unit+")"
	if (DimSize(data, 2) > 1)
		NoteText += " of layer "+num2str(layer)
	endif
	
	variable start	= (center -	sign(delta) * width/2 - offset)/delta + 0.5						// recalculate to rows / columns x2pnt()-like; the sign(delta) reverses the direction for negative delta values
	variable finish	= (center +	sign(delta) * width/2 - offset)/delta + 0.5						// shift the values by 0.5 to get the full rows /columns
//	start	= round(start/delta*1000)*delta/1000												// round the values depending on the delta size (to get rid of the floating point errors)
//	finish	= round(finish/delta*1000)*delta/1000
	start	= start < 0 ? 0 : start																// validate values
	finish	= finish > size ? size : finish
	
//	Print/D "Start initial:",start,"Finish initial:",finish										// debug
	
	variable beforeFrac = ceil(start) - start;		start  = ceil(start)						// fraction at the beginning of the full columns / rows
	variable afterFrac  = finish - trunc(finish);	finish = floor(finish)						// rest from the end of the range
	if (start > finish)																			// if range is smaller than one column
		beforeFrac += afterFrac - 1; afterFrac = 0
	endif
	width /= abs(delta)																			// normalize for scaled waves
	width = width == 0 ? 1 : width

//	Print/D "Overhead:",beforeFrac,"Start:",start,"Finish:",finish,"Rest:",afterFrac			// debug
	//++++++++++++++++++++++++++++++++++++ now integrate ++++++++++++++++++++++++++++++++++++++++
//	int timerRefNum = StartMSTimer																// debug
	
	ControlInfo/W=$pName MultiProfile;	int doMulti = (V_flag==2 && V_Value==1 && DimSize(data,2)>1 && WaveExists(multi))
	ControlInfo/W=$pName doZProfiles;	int doZPro  = (V_flag==2 && V_Value==1 && DimSize(data,2)>1 && WaveExists(multi))
	if (doMulti || doZPro)																		// do the summing for all layers
		if (mode)
			if (start < finish)																	// make sure there is actually something to extract
				Duplicate/Free/RMD=[start,finish-1][][] data, tempWave
				MatrixOP/Free tempSum = sumCols(tempWave)
			else
				Duplicate/Free/RMD=[0][][] data, tempSum
				WaveTransform/O setZero tempSum
			endif
			MatrixOP/Free tempWave = transposeVol(tempSum,4)
		else
			if (start < finish)
				Duplicate/Free/RMD=[][start,finish-1][] data, tempWave
				MatrixOP/Free tempSum = sumRows(tempWave)
			else
				Duplicate/Free/RMD=[][0][] data, tempSum
				WaveTransform/O setZero tempSum
			endif
			MatrixOP/Free tempWave = transposeVol(tempSum,1)
		endif
		
		Redimension/D tempWave;	Duplicate/O tempWave, multi
		if (beforeFrac > 0)
			multi += (mode == 0 ? data[p][start-1][q] : data[start-1][p][q]) * beforeFrac		// part of first line
		endif
		if (afterFrac > 0)
			multi += (mode == 0 ? data[p][finish][q]  : data[finish][p][q]) * afterFrac			// rest of the last line
		endif
		multi /= width																			// normalize to the integrated width
	else
		Redimension/D/N=(DimSize(data, mode)) secL;	secL = 0
		if (beforeFrac > 0)
			secL += (mode == 0 ? data[p][start-1][layer] : data[start-1][p][layer]) * beforeFrac
		endif
		for (i = start; i < finish; i++)
			secL += (mode == 0 ? data[p][i][layer]		  : data[i][p][layer])
		endfor
		if (afterFrac > 0)
			secL += (mode == 0 ? data[p][finish][layer]  : data[finish][p][layer]) * afterFrac
		endif
		secL /= width
	endif
	
	//+++++++++++++++++++++++++++++++++++++++ normalize profile ++++++++++++++++++++++++++++++++++
	ControlInfo/W=$pName Normto1;	int doNorm = V_Value
	if (doNorm)																					// normalizes the spectrum to the range 0-1
		if (doMulti || doZPro)
			for (i = 0; i < DimSize(multi, 1); i++)
				Duplicate/Free/R=[][i] multi, tempCol
				WaveStats/Q tempCol
				start = V_min; delta = V_max-start
				delta = delta == 0 ? 1 : delta
				multi[][i] = (multi-start)/delta
			endfor
		else
			WaveStats/Q secL
			start = V_min; delta = V_max-start
			delta = delta == 0 ? 1 : delta
			MatrixOP/O secL=(secL-start)/delta
		endif
	endif
	
	if (doMulti || doZPro)																		// set the scale last, after MatrixOP was executed
		SetScale/P x, DimOffset(data,mode), DimDelta(data,mode), WaveUnits(data,mode), multi
		SetScale/P y, DimOffset(data,2), 	DimDelta(data,2), 	 WaveUnits(data,2),	   multi
		SetScale d,0,0, WaveUnits(data,-1), multi
		Note/K multi, NoteText																	// write comments
		Duplicate/O/R=[][layer] multi, secL														// just copy out the respective layer data
		Redimension/N=(-1,0) secL
		if (doZPro)
			ControlInfo/W=$pName doZProfMode
			int zProf_Mode = WhichListItem(S_Value, kZProfileModes);	zProf_Mode = zProf_Mode < 0 ? 0 : zProf_Mode
			Switch(zProf_Mode)
				case 0:
					MatrixOP/O stg:SectionLine_Z = sumRows(multi^t)/numRows(multi)
				break
				case 1:
					MatrixOP/O stg:SectionLine_Z = maxRows(multi^t)
				break
				case 2:
					MatrixOP/O stg:SectionLine_Z = minRows(multi^t)
				break
			EndSwitch
			SetScale/P x, DimOffset(data,2), DimDelta(data,2), WaveUnits(data,2), stg:SectionLine_Z
			SetScale/P d, 0, 1, WaveUnits(data,-1), stg:SectionLine_Z
		endif
	else
		SetScale/P x, DimOffset(data,mode), DimDelta(data,mode), WaveUnits(data,mode), secL
		SetScale d,0,0, WaveUnits(data,-1), secL												// add data units
		Note/K secL, NoteText
	endif
	
	//+++++++++++++++++++++++++++++++++ prepare loaded profile +++++++++++++++++++++++++++++++++++
	string loadStr = "LoadProfile_"+SelectString(mode,"Horiz","Vert")							// new in ver. 4.45
	SVAR/Z srcPath = stg:$("Load"+SelectString(mode,"Horiz","Vert")+"WaveLoc")
	wave/Z loadPro = stg:$loadStr
	if (WaveExists(loadPro))
		WaveTransform/O/V=(NaN) setConstant loadPro												// clear old data
	endif
	
	if (SVAR_Exists(srcPath))																	// prepare profile from external wave
		wave/Z loadSrc = $srcPath
		if (WaveExists(loadSrc))
			Duplicate/O secL, stg:$loadStr
			wave loadPro = stg:$loadStr
			WaveTransform/O/V=(NaN) setConstant loadPro
			[start,finish] = findOverlap(loadPro, loadSrc)
			if (!numtype(start+finish))
				loadPro[start,finish] = loadSrc(x)
				if (doNorm)
					WaveStats/Q loadPro
					start = V_min; delta = V_max-start
					delta = delta == 0 ? 1 : delta
					MatrixOP/O loadPro=(loadPro-start)/delta
				endif
			endif
		endif
	endif
	
//	Print StopMSTimer(timerRefNum)/1000000, "seconds"											// debug
	
	Profile_UpdateLabelDisplay(pName, gName)
	Profile_ShiftMultiProfile(pName)															// update multiple profiles
	return 0
end

//################################################################################################

static function [variable p1, variable p2] findOverlap(wave w1, wave w2)
	p1 = NaN; p2 = NaN
	variable lx, rx, tmp, dx1 = DimDelta(w1,0), dx2 = DimDelta(w2,0)
	int dim1 = DimSize(w1,0)-1, dim2 = DimSize(w2,0)-1
	int rev1 = sign(dx1)==-1, rev2 = sign(dx2)==-1												// mirror for negative delta values => reverses assignment
	lx = rev1 ? pnt2x(w1,dim1) : pnt2x(w1,0)
	rx = rev2 ? pnt2x(w2,dim2) : pnt2x(w2,0)
	variable left = lx > rx ? lx : rx															// biggest value  => left border
	lx = rev1 ? pnt2x(w1,0) : pnt2x(w1,dim1)
	rx = rev2 ? pnt2x(w2,0) : pnt2x(w2,dim2)
	variable right = lx < rx ? lx : rx															// smallest value => right border
	if (right < left)
		return [p1, p2]																			// no overlap
	endif
	lx = rev1 ? right : left
	rx = rev1 ? left : right
	p1 = x2pnt(w1,lx) + (pnt2x(w1,x2pnt(w1,lx))<(rev2 ? pnt2x(w2,dim2) : pnt2x(w2,0)))			// make sure the range does not exceed the input
	p2 = x2pnt(w1,rx) - (pnt2x(w1,x2pnt(w1,rx))>(rev2 ? pnt2x(w2,0) : pnt2x(w2,dim2)))
	return [p1, p2]
end

//################################################################################################

static function Profile_UpdateLabelDisplay(string pName, string gName)							// creates an annotation which shows data labels
	if (!kShowLabelDisplay)
		return 0
	endif
	
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc
	int  mode = para[0], found
		
	ControlInfo/W=$pName SetLayer; int layer = V_Value
	
	string str = "", tmp = ""
	if (DimSize(data,2) > 1)
		if (DimOffset(data,2) != 0 || DimDelta(data,2) != 1)									// scaled layer position
			string unit = SelectString(strlen(WaveUnits(data,2)),""," ") + WaveUnits(data,2)
			str = num2str(DimOffset(data,2)+layer*DimDelta(data,2)) + unit
		endif
		tmp = GetDimLabel(data, 2, layer)	 													// the layer label
	endif
	variable pntPos	= round((para[1+mode*2] - DimOffset(data,1-mode))/DimDelta(data,1-mode))	// current center position converted to points
	pntPos = limit(pntPos, 0, DimSize(data,1-mode)-1)
	str += SelectString(strlen(str)>0 && strlen(tmp)>0,"","\r") + tmp							// add line break
	tmp  = GetDimLabel(data, 1-mode, pntPos)													// row / column labels
	str += SelectString(strlen(str)>0 && strlen(tmp)>0,"","\r") + tmp
	
	found = FindListItem("ImageLabelDisplay", AnnotationList(gName))>-1
	if((strlen(str) && !found) || found)
		TextBox/W=$gName/C/N=ImageLabelDisplay/F=0/A=RT SelectString(strlen(str),"","\\Zr075")+str
	endif
	
	// ++++++++++++++++++++++++++++++ external profile +++++++++++++++++++++++++++++++++++++++++++
	string scrName = "", loadThis = SelectString(mode,"Horiz","Vert")							// new in ver. 4.45
	SVAR/Z scrPath = stg:$("Load"+loadThis+"WaveLoc")
	wave/Z loadPr  = stg:$("LoadProfile_"+loadThis)
	if (SVAR_Exists(scrPath))
		scrName = ParseFilePath(0, scrPath, ":", 1, 0)
	endif
	if (WaveExists(loadPr)&&strlen(scrName))
		WaveStats/Q loadPr
		str = "\\Zr075Loaded Profile: \s(LoadProfile_"+loadThis+") \\K(34000,10000,60000)"+scrName+SelectString(V_npnts,"\\K(65355,0,0) \f01(no overlap!)","")
	endif
	found = FindListItem("LoadProfileDisplay", AnnotationList(gName))>-1
	if((strlen(str) && !found) || found)
		TextBox/W=$gName/C/N=LoadProfileDisplay/X=(mode?6:8)/Y=1/E=2/F=0/A=$SelectString(mode,"LB","LT") str
	endif
	return 0
end

//################################################################################################

static function Profile_CheckBordersAndUpdateAll(string pName)									// check if the lines are within the boundaries, if not restrict
	checkBorderForDim(pName, 0)
	checkBorderForDim(pName, 1)
	Profile_UpdateLinesAndProfile(pName)														// set the lines to the restricted values and recalculates result
	return 0
end
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function checkBorderForDim(string pName, int dim)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc

	variable off = DimOffset(data,dim), del = DimDelta(data,dim), size = DimSize(data,dim)
	variable pos = para[3-2*dim], width = para[4-2*dim], halfShft = 0
	ControlInfo/W=$pName SnapToGrid
	if (V_Value)
		width	 = del * round(width/del)
		halfShft = del/2 *!mod(width/del,2)														// for even widths the center needs to be shifted over by half a row/col
		pos		 = off+halfShft + del*round((pos-(off+halfShft))/del)
	endif
	
	variable boundary1 = off + (width*sign(del) - del)/2
	variable boundary2 = off - (width*sign(del) - del)/2 + del*(size-1)
	pos = (pos-boundary1)*sign(del) < 0 ? boundary1 : pos										// too low?
	pos = (pos-boundary2)*sign(del) > 0 ? boundary2 : pos										// too high?
	width = width > abs(del*size) ? abs(del*size) : width										// too big?
	para[3-2*dim] = pos																			// return correct value
	para[4-2*dim] = width
end

//#################################support functions for control elements#########################

function Profile_PopupControl(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)
		Profile_CheckBordersAndUpdateAll(s.win)
	endif
	return 0
end

//################################################################################################

function Profile_DisplayCheckBox(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		Profile_UpdateImageContrast(s.win, keepRange)
	endif
	return 0
end

//################################################################################################

function ModeCheckBoxControl(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		Profile_ExecuteCheckboxAction(s.win, s.ctrlName,s.checked)
	endif
	return 0
end
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function Profile_ExecuteCheckboxAction(string pName, string type, int checked)
	DFREF stg = getWorkFolder(pName)
	wave para = stg:CurrentLines
	string gName = getGraph(pName)
	string sPath = gname+"#floatPanel#zProfile"
	
	StrSwitch (type)
		case "IntColCheckBox":
			para[0] = 0
			KillCompareLine(pName)																// delete snapshot when switching modes
			Profile_AppendSectionTrace(pName)
			CheckBox IntRowCheckBox ,win=$pName ,value = 0
			ControlInfo/W=$pName NormModeTitle													// new in ver. 4.45
			if (V_flag == 10)
				TitleBox NormModeTitle ,win=$pName ,title="Each Col.:"
			endif
//			if (WinType(sPath))
//				ModifyGraph/W=$sPath swapXY=!para[0]
//			endif
		break
		case "IntRowCheckBox":
			para[0] = 1
			KillCompareLine(pName)
			Profile_AppendSectionTrace(pName)
			CheckBox IntColCheckBox ,win=$pName ,value = 0
			ControlInfo/W=$pName NormModeTitle
			if (V_flag == 10)
				TitleBox NormModeTitle ,win=$pName ,title="Each Row:"
			endif
//			if (WinType(sPath))
//				ModifyGraph/W=$sPath swapXY=!para[0]
//			endif
		break
		case "FullWidth":
			Profile_SetMinFullLineValues(pName,1,para[0],checked)
		break
		case "MinWidth":
			Profile_SetMinFullLineValues(pName,0,para[0],checked)
		break
		case "toggleNormto1":
			ControlInfo/W=$pName Normto1
			CheckBox Normto1 ,win=$pName ,value = 1-V_Value
		break
		case "toggleMousetoSpec":
			ControlInfo/W=$pName MousetoSpec
			CheckBox MousetoSpec ,win=$pName ,value = 1-V_Value; checked = 1-V_Value
		case "MousetoSpec":
			if (!checked)																		// make sure to remove the profiles properly
				RemoveFromGraph/Z/W=$gName MouseProfile
			if (WinType(sPath))
				RemoveFromGraph/Z/W=$sPath MouseProfile
			endif
			endif
		break
		case "toggleSnapToGrid":
			ControlInfo/W=$pName SnapToGrid
			CheckBox SnapToGrid	 ,win=$pName ,value = 1-V_Value
		case "SnapToGrid":
			Profile_SetPosWidthControls(pName)
		break
		case "MultiProfile":
			Profile_AppendSectionTrace(pName)
			Profile_UpdateLinesAndProfile(pName)												// now create the profiles
		break
		case "doZProfiles":
			SetActiveSubwindow $gName
			KillWindow/Z #floatPanel
			if (!checked)
				break
			endif
			
			ControlInfo/W=$pName SetLayer;	int layer = abs(V_flag)==5 ? V_Value : 0
			if (!WaveExists(stg:SectionLine_Z))
				Make/D/O/N=2 stg:SectionLine_Z = 0
				layer = 0
			endif
			
			GetWindow $gName, expand
			variable gExpand = V_value;	gExpand = gExpand == 0 ? 1 : gExpand					// pixel conversion and expansion factor
			variable pixConv = ScreenResolution/72*gExpand
			GetWindow $gName psizeDC
			variable sizeW = max((V_right-V_left)/4,subPanelMin)/pixConv
			variable sizeH = max((V_bottom-V_top)/4,subPanelMin)/pixConv
			variable wL = (V_left+20)/pixConv, wT = (V_top+20)/pixConv, wR = wL+sizeW, wB = wT+sizeH
			
			string prevStruct = GetUserData(pName,"","dragStruct")								// load previous position
			if (strlen(prevStruct))
				STRUCT DragStruct d
				StructGet/S d, prevStruct
				wL = d.subwRect.left/pixConv
				wT = d.subwRect.top/pixConv
				wR = d.subwRect.right/pixConv
				wB = d.subwRect.bottom/pixConv
			endif
			
			NewPanel/HOST=$gName/N=floatPanel/W=(wL,wT,wR,wB)
			Display/HOST=$(gName+"#floatPanel")/FG=(FL,FT,FR,FB)/N=zProfile stg:SectionLine_Z
			ModifyGraph/W=$sPath standoff=0, minor=1, mirror=2, ZisZ=1, frameStyle=1, swapXY=1 	//,swapXY=!para[0]
			SetAxis/W=$sPath/A/E=1/N=1 bottom
			Cursor/W=$sPath/P/S=2/H=3/L=1/T=2/C=(34000,10000,60000) D, SectionLine_Z, layer		// layer indicator
			SetActiveSubwindow $gName
		break
	EndSwitch
	
	Profile_CheckBordersAndUpdateAll(pName)														// check if the lines go over the borders of the wave
	Profile_SetFullMinCheckbox(pName)															// toggle check-boxes
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function KillCompareLine(string pName)
	DFREF stg = getWorkFolder(pName)
	RemoveFromGraph/W=$getGraph(pName)/Z CompareLine
	KillWaves/Z stg:CompareLine
end

//################################################################################################

static function Profile_SetMinFullLineValues(string pName, int type, int mode, int activate)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc

	variable size	= DimSize(data,	 1-mode)
	variable offset	= DimOffset(data,1-mode)	
	variable delta  = DimDelta(data, 1-mode)
	variable init = abs(delta)*round((size-1) / 25)												// calculate ca. 5% of width
	init = init <= abs(delta) ? 2*abs(delta) : init												// make sure that's not too small
	
	if (type)																					// 1: full width, 0: min width
		offset += size<100 ? delta*floor((size-1)/2) : delta*(size-1)/2							// align to grid for small waves
		offset = abs(offset) < 10^-10*init ? 0 : offset											// properly center around zero (rounding errors)
		para[1+mode*2] = offset																	// center position
		para[2+mode*2] = activate ? abs(delta)*size : init										// full with or ca. 20 %
	else
		para[2+mode*2] = activate ? abs(delta) : init											// one line or 20 %
	endif
	return 0
end

//################################################################################################

function ChangeParameterControl(STRUCT WMSetVariableAction &s) : SetVariableControl				// execute variable changes and sets the check-boxes
	if (s.eventCode == 1 || s.eventCode == 2)
		Profile_ShiftMultiProfile(s.win)														// need to shift the traces first before recalculation takes place
		if (CmpStr(s.ctrlName,"SetLayer") == 0)
			ModifyImage/W=$getGraph(s.win) ''#0 ,plane = s.dval									// replace the currently displayed layer
			Profile_UpdateImageContrast(s.win, resetRange)										// updates the contrast sliders for the new plane
//			Profile_UpdateLinesAndProfile(s.win)												// triggers profile recalculation
		endif
		Profile_CheckBordersAndUpdateAll(s.win)
		Profile_SetFullMinCheckbox(s.win)
	endif
	return 0
end
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function Profile_SetFullMinCheckbox(string pName)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc
	int  mode = para[0]
	variable delta = abs(DimDelta(data,1-mode)), size = DimSize(data,1-mode)
	CheckBox FullWidth ,win=$pName ,value = (para[2+mode*2] >= delta*size)
	CheckBox MinWidth  ,win=$pName ,value = (para[2+mode*2] <= delta)
	return 0
end

//################################################################################################

function Profile_ContrastSliderProc(STRUCT WMSliderAction &s) : SliderControl
	if (s.eventCode & 1)
		ControlInfo/W=$s.win ContrastLow;	variable minVal = V_Value
		ControlInfo/W=$s.win ContrastHigh;	variable maxVal = V_Value
		ControlInfo/W=$s.win ContrastColor;	string color = S_Value
		ModifyImage/W=$getGraph(s.win) ''#0 ctab={minVal,maxVal,$color,0}
	endif
	return 0
end

//################################################################################################

function Profile_ContrastColorProc(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)
		ControlInfo/W=$s.win ContrastLow;	variable minVal = V_Value
		ControlInfo/W=$s.win ContrastHigh;	variable maxVal = V_Value
		ModifyImage/W=$getGraph(s.win) ''#0 ctab={minVal,maxVal,$s.popStr,0}
	endif
	return 0
end

//################################################################################################

function Profile_MainButtonProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		Profile_ExecuteButtonCommand(s.win, s.ctrlName)
	endif
	return 0
end

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static Constant dontUpdate = 0
static Constant keepRange = 1
static Constant resetRange = 2

static function Profile_ExecuteButtonCommand(string pName, string bAction)
	string gName = getGraph(pName)
	DFREF stg = getWorkFolder(pName)
	if (DataFolderRefStatus(stg) == 0)															// something is wrong if the data is not found => exit
		return -1
	endif
	
	SVAR wLoc = stg:InwaveLocation
	wave secL = stg:SectionLine
	wave para = stg:CurrentLines
	wave data = $wLoc
	int  mode = para[0]
	
	string start = num2str(ChangePrecision(para[1+mode*2]-para[2+mode*2]/2, 4))
	string stop  = num2str(ChangePrecision(para[1+mode*2]+para[2+mode*2]/2, 4))
	string units = WaveUnits(data,1-mode); units[0] = SelectString(strlen(units),""," ")
	ControlInfo/W=$pName SetLayer;	int layer = abs(V_flag)==5 ? V_Value : 0
	
	int i, doQuit = 0
	int direction = 1																			// shift / rotation direction starts positive
	int updateContrasts = 0																		// updates the contrast sliders (1 = don't change range, 2 = reset range)
	int updateLayerCtrl = 0																		// updates layer-related controls
	variable intensitySum = 0																	// for later re-normalization
	string logStr = ""
	
	Strswitch (bAction)
		case "bSaveQuit":
			logStr = "Session closed."
		case "bQuit":
			doQuit = 1
		break
		case "bSaveSpec":
			logStr = Profile_SaveProfile(pName)
		break
		case "bSaveAsNew":
			do
				string newName = UniqueName(NameOfWave(data), 1, 0)
				Prompt newName, "Enter a new name:"
				DoPrompt/Help="The data will be saved in the current folder. Any exists data with the same name will be overwritten." "Save Edits with a New Name", newName
				if (V_flag)
					break
				endif
				if (!strlen(newName))
					DoAlert/T="Error when saving data" 0, "The new name cannot be empty."
					continue
				endif
				if (!CmpStr(GetDataFolder(1)+PossiblyQuoteName(newName), wLoc))
					DoAlert/T="Name and location are identical" 0, "You cannot overwrite the original data with this functionality. Please choose a different name and / or location."
					continue
				endif
				if (WaveExists($newName))
					DoAlert/T="Name conflict" 2, "A wave with that name already exists. Do you want to overwrite?"
					if (V_flag == 2)
						continue
					elseif (V_flag == 3)
						break
					endif
				endif
				
				Duplicate/O	data, $newName /wave=newdata
				ReplaceWave/W=$gName image=$PossiblyQuoteName(NameOfWave(data)), newdata		// replace displayed image and title
				
				GetWindow $gName wtitle
				DoWindow/T $gName,ReplaceString(NameOfWave(data), S_value, newName)				// change the title
				Profile_UndoLastUpdate(pName, "restore", bAction)								// write the backup into original input
				Duplicate/O	newdata, stg:StartCopy												// now replace the start copy
				wLoc = GetWavesDataFolder(newdata,2)											// point to the new location
				Profile_AppendSectionTrace(pName)												// update the color scale
				logStr = "Saved under a new name: "+ newName
				break
			while(1)
		break
		case "bCompare":
			Duplicate/O secL, stg:CompareLine													// duplicate current profile for comparison
			Profile_AppendSectionTrace(pName)
			logStr  = "Made snapshot of " + SelectString(mode,"horizontal","vertical")
			logStr += " profile (Range: "+start+"-"+stop+units+")" + SelectString(WaveDims(data)>2,""," of layer "+num2str(layer)) + "."
		break
		case "bUndoAll":
			Profile_UndoLastUpdate(pName, "all", bAction)
			logStr = "All changes undone."
			updateContrasts = resetRange
		break
		case "bUndoLast":
			Profile_UndoLastUpdate(pName, "last", bAction)
			logStr = "Last change undone."
			updateContrasts = resetRange
		break
		case "bAlign":
			logStr = Profile_Align(pName)														// no Profile_UndoLastUpdate() here => has its own save functionality
			updateContrasts = keepRange
		break
		case "bDeleteLayer":
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = "Deleted layer "+num2str(layer+1)+" of "+num2str(DimSize(data,2))+"."
			updateLayerCtrl = 1
			DeletePoints/M=2 layer,1, data
			if (DimSize(data,2) == 1)
				Redimension/N=(-1,-1,0) data													// convert to a proper 2D image
			endif
		break
		case "bSumLayers":																		// merges 3D layers into an 2D image
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = "Merged all layers into 2D image."
			updateLayerCtrl = 1
			updateContrasts = resetRange
			
			MatrixOP/Free tMat = sumBeams(data)/waveLayers(data)								// merge layers
			CopyScales data, tMat;	Note tMat, note(data)										// copy scales and notes over
			Duplicate/O tMat, data
		break
		case "bNormalize":																		// normalize to profile shape
			WaveStats/Q secL;	intensitySum = V_avg
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_DivideByProfile(pName)
			updateContrasts = resetRange
		break
		case "bNormArea":
		case "bNormMax":																		// normalize to maximum or area
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_DivideByValue(pName, bAction, intensitySum)
			updateContrasts = resetRange
		break
		case "bNormUser":																		// normalize to measurement value
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = "Normalized by "+kUserInfoBtnName+"."
			updateContrasts = resetRange
			intensitySum = Profile_ExecuteUserInfoButton(data)
			Button bNormUser ,win=$pName ,disable= 2											// disable the button to avoid double clicking (more than once is meaningless anyway)			
		break
		case "bSubtractPr":	
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_SubtractProfile(pName)
			updateContrasts = keepRange
		break
		case "bRemoveSelection":
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_RemoveSelection(pName)
			updateContrasts = keepRange
		break
		case "bShiftYAxisMinus":
			direction = -1
		case "bShiftYAxisPlus":
			Profile_UndoLastUpdate(pName, "save", bAction)
			Profile_ShiftButtonStatus(pName, direction)											// enable and disable shift buttons
			logStr = Profile_ShiftImage(pName, direction)
			updateContrasts = keepRange
		break
		case "bLayer_RotX_neg":
			direction = -1
		case "bLayer_RotX_pos":
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_RotateImage(data, stg:LastCopy, gName, "X", direction)
			updateLayerCtrl = 1
			updateContrasts = resetRange
		break
		case "bLayer_RotY_neg":
			direction = -1
		case "bLayer_RotY_pos":
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_RotateImage(data, stg:LastCopy, gName, "Y", direction)
			updateLayerCtrl = 1
			updateContrasts = resetRange
		break
		case "bLayer_RotZ_neg":
			direction = -1
		case "bLayer_RotZ_pos":
			Profile_UndoLastUpdate(pName, "save", bAction)
			logStr = Profile_RotateImage(data, stg:LastCopy, gName, "Z", direction)
			updateContrasts = keepRange
		break
	Endswitch
	
	ControlInfo/W=$pName MaintainInt
	if (V_Value && intensitySum != 0 && numtype(intensitySum) == 0)								// restore the average height of the spectrum (from the average height of the Profile)
		data[][][layer] *= intensitySum
		logStr += " Average intensity is maintained."
	endif
	
	ControlInfo/W=$pName cPrintEdits															// print logs if desired
	if (V_flag == 2 && V_Value == 1 && strlen(logStr))
		Print "[ImgPr v" + num2str(kVersion) + "] " + NameOfWave(data) + " (" + time() + "): " + logStr
	endif
	
	if (doQuit)
		Profile_QuitProgram(pName)
		return 0
	endif
	
	if (updateLayerCtrl)
		Profile_UpdateLayerControls(pName)
	endif
	Profile_UpdateAxes(pName, StringMatch(bAction, "bLayer_*"))
	Profile_UpdateImageContrast(pName, updateContrasts)
	Profile_CheckBordersAndUpdateAll(pName)
	return 0
end

//################################################################################################

static function/S Profile_RotateImage(wave data, wave data_backup, string gName, string axis, variable direction)
	variable sX = DimSize(data,0)-1, sY = DimSize(data,1)-1, sZ = DimSize(data,2)-1				// values for 3D volume rotation
	variable oX = DimOffset(data,0), oY = DimOffset(data,1), oZ = DimOffset(data,2)
	variable dX = DimDelta(data,0),  dY = DimDelta(data,1),  dZ = DimDelta(data,2)
	string   uX = WaveUnits(data,0), uY = WaveUnits(data,1), uZ = WaveUnits(data,2)
	
	string axInfo = GetUserData(gName,"","axisFlip")
	variable lastX = NumberByKey("Xaxis",axInfo);	lastX = numtype(lastX) != 0 ? 1 : lastX		// saved axis flip for 3D volume rotation
	variable lastY = NumberByKey("Yaxis",axInfo);	lastY = numtype(lastY) != 0 ? 1 : lastY
	variable lastZ = NumberByKey("Zaxis",axInfo);	lastZ = numtype(lastZ) != 0 ? 1 : lastZ
	variable rotNo = NumberByKey("rotNo",axInfo);	rotNo = numtype(rotNo) != 0 ? 1 : rotNo		// number of rotations done
	
	int i
	StrSwitch (axis)
		case "X":
			if (direction == -1)			
				MatrixOP/Free tempVol = reverseRows(data);		MatrixOP/O data = transposeVol(tempVol,1)		// 90 deg rotation = flip and transpose
			else
				MatrixOP/Free tempVol = transposeVol(data,1);	MatrixOP/O data = reverseRows(tempVol)
			endif
			SetScale/P y, oZ + sZ*dZ*(direction == +1), sign(-direction)*dZ, uZ, data			// flip scales as well
			SetScale/P z, oY + sY*dY*(direction == -1), sign(+direction)*dY, uY, data
			for (i = 0; i < sZ+1; i++)
				SetDimLabel 1, i, $cleanStr(GetDimLabel(data_backup,2,((direction == -1) ? i : sZ-i)), ",|;|:|'|\"|", insert="_"), data
			endfor
			for (i = 0; i < sY+1; i++)															// copy labels into the right dimension
				SetDimLabel 2, i, $cleanStr(GetDimLabel(data_backup,1,((direction == +1) ? i : sY-i)), ",|;|:|'|\"|", insert="_"), data
			endfor
			SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(lastX)+";Yaxis:"+num2str(sign(-direction)*lastZ)+";Zaxis:"+num2str(sign(+direction)*lastY)+";rotNo:"+num2str(rotNo+1)+";"
		break
		case "Y":
			if (direction == -1)
				MatrixOP/Free tempVol = reverseCols(data);		MatrixOP/O data = transposeVol(tempVol,3)
			else
				MatrixOP/Free tempVol = transposeVol(data,3);	MatrixOP/O data = reverseCols(tempVol)
			endif
			SetScale/P x, oZ + sZ*dZ*(direction == +1), sign(-direction)*dZ, uZ, data
			SetScale/P z, oX + sX*dX*(direction == -1), sign(+direction)*dX, uX, data
			for (i = 0; i < sZ+1; i++)
				SetDimLabel 0, i, $cleanStr(GetDimLabel(data_backup,2,((direction == -1) ? i : sZ-i)), ",|;|:|'|\"|", insert="_"), data
			endfor
			for (i = 0; i < sX+1; i++)
				SetDimLabel 2, i, $cleanStr(GetDimLabel(data_backup,0,((direction == +1) ? i : sX-i)), ",|;|:|'|\"|", insert="_"), data
			endfor		
			SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*lastZ)+";Yaxis:"+num2str(lastY)+";Zaxis:"+num2str(sign(+direction)*lastX)+";rotNo:"+num2str(rotNo+1)+";"
		break
		case "Z":
			if (direction == -1)
				MatrixOP/Free tempVol = reverseCols(data)
				if (DimSize(data,2) > 1)
					MatrixOP/O data = transposeVol(tempVol,5)
				else
					MatrixOP/O data = tempVol^t													// z-axis rotation also works for 2D maps
				endif
			else
				if (DimSize(data,2) > 1)				
					MatrixOP/Free tempVol = transposeVol(data,5)
				else
					MatrixOP/Free tempVol = data^t
				endif
				MatrixOP/O data = reverseCols(tempVol)
			endif
			SetScale/P x, oY + sY*dY*(direction == +1), sign(-direction)*dY, uY, data
			SetScale/P y, oX + sX*dX*(direction == -1), sign(+direction)*dX, uX, data
			for (i = 0; i < sY+1; i++)
				SetDimLabel 0, i, $cleanStr(GetDimLabel(data_backup,1,((direction == -1) ? i : sY-i)), ",|;|:|'|\"|", insert="_"), data
			endfor
			for (i = 0; i < sX+1; i++)
				SetDimLabel 1, i, $cleanStr(GetDimLabel(data_backup,0,((direction == +1) ? i : sX-i)), ",|;|:|'|\"|", insert="_"), data
			endfor		
			SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*lastY)+";Yaxis:"+num2str(sign(+direction)*lastX)+";Zaxis:"+num2str(lastZ)+";rotNo:"+num2str(rotNo+1)+";"
		break
	EndSwitch
	
	return "Data rotated around "+axis+" in "+SelectString(direction+1,"clockwise","counter-clockwise")+" direction."
end

//################################################################################################

static function Profile_UpdateImageContrast(string pName, int updateMode)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation

	ControlInfo/W=$pName SetLayer;		int layer = abs(V_flag)==5 ? V_Value : 0
	ControlInfo/W=$pName cDoLogScale;	int doLog = V_flag==2 ? V_Value : 0
	ControlInfo/W=$pName ContrastColor
	if (updateMode == dontUpdate || abs(V_flag) != 3)
		return 0
	endif
	string color = S_Value
	
	WaveStats/Q/RMD=[][][layer] $wLoc
	variable wMin = V_min, wMax = V_max
	variable minV = V_min, maxV = V_max
	
	if (doLog)
		if (wMax <= 0)
			doLog = 0
			CheckBox cDoLogScale, win=$pName, value=0
		else
			wMin = wMin <= 0 ? wMax/10^6 : wMin
			minV = wMin
		endif
	endif
	
	if (updateMode == keepRange)																// mode 1 reads current setting from the contrast sliders
		ControlInfo/W=$pName ContrastLow;	minV = limit(V_Value,wMin,wMax)
		ControlInfo/W=$pName ContrastHigh;	maxV = limit(V_Value,wMin,wMax)
		minV = minV >= maxV ? wMin : minV
	endif
	
	string gName = getGraph(pName)
	string gAxis = StringFromList(0,ListMatch(AxisList(gName), "*2"))
	
	Slider ContrastLow	,win=$pName	,value=minV	,limits={wMin,wMax,0}
	Slider ContrastHigh	,win=$pName	,value=maxV	,limits={wMin,wMax,0}
	ModifyImage/W=$gName ''#0 ctab={minV,maxV,$color,0}, log=doLog
	ColorScale/W=$gName/C/N=ImageColorDisplay log=doLog
	ModifyGraph/W=$gName log($gAxis)=doLog
	return 0
end

//################################################################################################

static function Profile_UpdateAxes(string pName, int update)									// after rotation operation change axis direction and units																		
	if (!update)																				// [1] = do update; [-1] = undo rotation; [-2] = reset all
		return 0
	endif
	
	string gName = getGraph(pName)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave data = $wLoc
	SetScale d 0,0,WaveUnits(data,0), stg:LinesX												// the data units will be displayed on the x- and y-axes
	SetScale d 0,0,WaveUnits(data,1), stg:LinesY
	
	variable isR = NumberByKey("rotNo",GetUserData(gName,"","axisFlip"));	isR = numtype(isR) != 0 ? 1 : isR
	if (update == -2)																			// reset all axes to their initial orientation
		SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(DimDelta(data,0)))+";Yaxis:"+num2str(sign(DimDelta(data,1)))+";Zaxis:"+num2str(sign(DimDelta(data,2)))+";rotNo:0;"
	endif
	variable doX = NumberByKey("Xaxis",GetUserData(gName,"","axisFlip"));	doX = numtype(doX) != 0 ? 1 : doX
	variable doY = NumberByKey("Yaxis",GetUserData(gName,"","axisFlip"));	doY = numtype(doY) != 0 ? 1 : doY
	variable doZ = NumberByKey("Zaxis",GetUserData(gName,"","axisFlip"));	doZ = numtype(doZ) != 0 ? 1 : doZ
	
	string lastCmd = GetUserData(pName,"","lastAction")
	if (update == -1)																			// return axis direction (basically do the reverse operation)
		variable direction = -1
		StrSwitch (lastCmd)
			case "bLayer_RotX_neg":
				direction = 1
			case "bLayer_RotX_pos":
				SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(doX)+";Yaxis:"+num2str(sign(-direction)*doZ)+";Zaxis:"+num2str(sign(+direction)*doY)+";rotNo:"+num2str(isR-1)+";"
			break
			case "bLayer_RotY_neg":
				direction = 1
			case "bLayer_RotY_pos":
				SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*doZ)+";Yaxis:"+num2str(doY)+";Zaxis:"+num2str(sign(+direction)*doX)+";rotNo:"+num2str(isR-1)+";"
			break
			case "bLayer_RotZ_neg":
				direction = 1
			case "bLayer_RotZ_pos":
				SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*doY)+";Yaxis:"+num2str(sign(+direction)*doX)+";Zaxis:"+num2str(doZ)+";rotNo:"+num2str(isR-1)+";"
			break
		EndSwitch
		doX = NumberByKey("Xaxis",GetUserData(gName,"","axisFlip"))
		doY = NumberByKey("Yaxis",GetUserData(gName,"","axisFlip"))
	endif
	
	if (doX == -1)
		SetAxis/W=$gName/A/R bottom
	else
		SetAxis/W=$gName/A bottom
	endif
	if (doY == -1)
		SetAxis/W=$gName/A/R left
	else
		SetAxis/W=$gName/A left
	endif
	
	if (update>0 || (update==-1 && StringMatch(lastCmd,"bLayer_*")) || (update==-2 && isR))		// reset center and width to adapt to new range when the data seems to be rotated
		Profile_SetMinFullLineValues(pName,1,0,0)
		Profile_SetMinFullLineValues(pName,1,1,0)
		KillCompareLine(pName)																	// killing the wave prevents re-appendage
	endif
	return 0
end

//################################################################################################

static function/S Profile_SaveProfile(string pName)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave   para = stg:CurrentLines
	wave   secL = stg:SectionLine
	wave/Z multi= stg:MultiSection
	wave   data = $wLoc
	variable mode = para[0]
	
	string from	= num2str(ChangePrecision(para[1 + mode * 2] - para[2 + mode * 2] / 2,4))		// current range
	string to	= num2str(ChangePrecision(para[1 + mode * 2] + para[2 + mode * 2] / 2,4))
	int i, is3D = DimSize(data,2) > 1
	
	variable Opt_cutName  = (GetkeyState(0) & 2^2) != 0											// keyboard shortcut ('shift') to just cut the name from the ending
	variable Opt_sourceDF = (GetkeyState(0) & 2^1) != 0											// keyboard shortcut ('alt') to save the output to the source's path
	
	string outName, unit
	string outPath = SelectString(Opt_sourceDF,"",GetWavesDataFolder(data, 1))
	ControlInfo/W=$pName FullWidth
	if (V_Value || Opt_cutName)
		outName	= NameOfWave(data)+"_pr"+SelectString(mode,"H","V")
	else
		outName = NameOfWave(data)+SelectString(mode,"_H(" ,"_V(")+from+" to "+to+")"
	endif
	
	ControlInfo/W=$pName MultiProfile															// save the multi-profile
	if (V_flag==2 && V_Value==1 && is3D && WaveExists(multi))
		unit = WaveUnits(data,2);	unit[0] = SelectString(strlen(unit),""," ")
		variable pnt = round((para[1+mode*2] - DimOffset(data,1-mode))/DimDelta(data,1-mode))	// current center position converted to points
		string dimlbl = "", lbl = ""
		for (i = 0; i < DimSize(multi,1); i++)
			if (DimOffset(data,2) != 0 && DimDelta(data,2) != 1)
				lbl = num2str(DimOffset(data,2)+i*DimDelta(data,2))+unit						// add layer info as column labels
			endif
			dimlbl = GetDimLabel(data, 2, i)
			lbl   += SelectString(strlen(lbl)>0 && strlen(dimlbl)>0,""," | ") + dimlbl
			dimlbl = GetDimLabel(data, 1-mode, pnt)												// row / column labels
			lbl   += SelectString(strlen(lbl)>0 && strlen(dimlbl)>0,""," | ") + dimlbl
			SetDimLabel 1, i, $lbl, multi
		endfor
		Duplicate/O multi, $(outPath+SelectString(Opt_sourceDF, outName+"_map", PossiblyQuoteName(outName+"_map")))
	endif
	
	ControlInfo/W=$pName SetLayer
	if (is3D && abs(V_flag) == 5)
		sprintf outName, "%s_L%.2d", outName, V_Value
	endif
	Duplicate/O secL, $(outPath+SelectString(Opt_sourceDF, outName, PossiblyQuoteName(outName)))
	
	unit = WaveUnits(data,1-mode);	unit[0] = SelectString(strlen(unit),""," ")
	return "Saved "+SelectString(mode,"horizontal","vertical")+" profile (integrated from " + from+unit+" to "+to+unit+") => output to "+outName+"."
end
//++++++++++++++++++++++++++++++++++++++++ number rounding +++++++++++++++++++++++++++++++++++++++
static function ChangePrecision(variable value, variable numSigDigits)							// round values to select-able number of digits
	string str;	sprintf str, "%.*g\r", numSigDigits, value
	return str2num(str)
end

//################################################################################################

static function/S Profile_RemoveSelection(string pName)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc
	int  mode = para[0]
	
	variable offset = DimOffset(data,1-mode), delta = DimDelta(data,1-mode), size = DimSize(data,1-mode)
	variable start	= para[1 + mode * 2] - sign(delta)*para[2 + mode * 2] / 2					// calculate the position of the lines (reverse if sign is flipped)
	variable stop	= para[1 + mode * 2] + sign(delta)*para[2 + mode * 2] / 2
	variable newOff	= offset-(start-stop)														// recalculate the offset (is used if the removed section is at the left / bottom side)
	
//	Print "start:",start,"stop:", stop, "offset:", newOff										// debug (range in wave scales)
	start = round((start + delta/2 - offset)/delta)												// which points do the lines include
	stop  = round((stop  + delta/2 - offset)/delta)
//	Print "start:",start,"stop:", stop, "count:", abs(stop - start)								// debug (range in wave points)
	
	if (abs(stop-start) >= size)
		Abort "Range too big. Cannot delete the whole data."
	endif
	
	DeletePoints/M=(1-mode) start, abs(stop-start), data										// delete the selected points
	if (start == 0)																				// shifts the scale only if the wave is cut at the old offset
		if (mode)
			SetScale/P x newOff, delta, WaveUnits(data,1-mode),  data
		else
			SetScale/P y newOff, delta, WaveUnits(data,1-mode),  data
		endif
	endif
	return SelectString(mode,"Horizontal","Vertical")+" ROI deleted: " + num2str(abs(stop-start))+" (range "+num2str(start)+"-"+num2str(stop)+") of "+num2str(size)+" lines."
end

//################################################################################################

static function/S Profile_ShiftImage(string pName, variable shiftDir)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave data = $wLoc
	wave/Z lastcpy = stg:LastCopy																// read wave data from the copy if available
	if (!WaveExists(lastcpy))
		wave lastcpy = $wLoc
	endif
	
	variable i, modkey = (GetKeyState(0) & 2^0) != 0											// when the 'ctrl/cmd' key is pressed, the excess amount will be cut from the front of the spectrum	
	variable rows = DimSize(lastcpy,0),  cols = DimSize(lastcpy,1)
	variable xdel = DimDelta(lastcpy,0), xoff = DimOffset(lastcpy,0)
	variable ydel = DimDelta(lastcpy,1), yoff = DimOffset(lastcpy,1)
	shiftDir *= sign(xdel)																		// shift in the opposite direction for inverted data
	
	variable addrow = ceil((cols-1)*abs(ydel/xdel))												// how many rows need to be added for the shift
	variable newOff	= xoff + shiftDir*yoff														// the new x-offset
	newOff += sign(ydel)<0 ? shiftDir*(cols-1)*ydel : 0											// for negative y-deltas add the full width (basically start from the bottom)
	if (modkey && (addrow < rows-10))															// delete incomplete parts
		Redimension/D/N=(rows-addrow,-1,-1) data
		newOff += shiftDir ==  1*sign(xdel) ? addrow*xdel : 0
	else
		Redimension/D/N=(rows+addrow,-1,-1) data
		newOff -= shiftDir == -1*sign(xdel) ? sign(xdel)*(cols-1)*abs(ydel) : 0					// for negative shifts the start offset needs to be shifted as well (by the y-width of the spectrum)
	endif
	SetScale/P x newOff, xdel, WaveUnits(data,0), data
	
	if (WaveDims(data) == 3)
		MultiThread data = interp3D(lastcpy, newOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel, z)
	else
		MultiThread data = interp2D(lastcpy, newOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel)
	endif
	MatrixOP/Free tMat=replaceNaNs(data,0)														// fill NaNs with zero
	data = tMat
	
	//++++++++++++++++++++++++++ if there are zero rows, remove them +++++++++++++++++++++++++++++
	for (i = 0; i < DimSize(data,0); i++)
		MatrixOp/Free partSum = Sum(Row(data,i))
		if (partSum[0] != 0)
			break
		endif
	endfor
	if (i > 0)
		DeletePoints 0,i, data
		SetScale/P x DimOffset(data,0)+i*DimDelta(data,0), DimDelta(data,0), WaveUnits(data,0),  data
	endif
	
	variable excessRows = 0, lastData = -1
	for (i = DimSize(data,0)-1; i > 0; i -= 1)
		MatrixOp/Free partSum = Sum(Row(data,i))
		if (partSum[0] == 0)
			excessRows += 1
		else
			if (lastData == -1)
				lastData = i
			endif
		endif
	endfor
	if (excessRows > 0 && lastData != -1)
		DeletePoints lastData+1, excessRows, data
	endif
	
	return "Data shifted to "+SelectString(shiftDir+1,"negative","positive")+" y-axis values."
end

//################################################################################################

static function Profile_ShiftButtonStatus(string pName, int mode)								// mode: -1 = -Y; 0 = undo; 1 = +Y;
	int dState = 0
	ControlInfo/W=$pName bShiftYAxisPlus;	dState = (V_disable == 2) ?  1 : dState				// first, find the current disable state of the two buttons
	ControlInfo/W=$pName bShiftYAxisMinus;	dState = (V_disable == 2) ? -1 : dState
	Button bShiftYAxisPlus  ,win=$pName ,disable=(dState == 0 && mode ==  1 ? 2 : 0)			// deactivate the relevant button
	Button bShiftYAxisMinus ,win=$pName ,disable=(dState == 0 && mode == -1 ? 2 : 0)
	Button bUndoLast		,win=$pName ,userdata(lastShift)=num2str(mode)
end

//################################################################################################

static function/S Profile_SubtractProfile(string pName)
	DFREF stg = getWorkFolder(pName)
	SVAR   wLoc = stg:InwaveLocation
	wave   secL = stg:SectionLine
	wave/Z cmpL	= stg:CompareLine
	wave   para = stg:CurrentLines
	wave   data = $wLoc
	
	ControlInfo/W=$pName SubtractScale;	variable scale = V_flag==5 ? V_Value : 1				// profile scaling (added in ver. 4.20)
	ControlInfo/W=$pName SetLayer;		int layer = abs(V_flag)==5 ? V_Value : 0
	
	int i, mode = para[0]
	string start  = num2str(ChangePrecision(para[1+mode*2]-para[2+mode*2]/2, 4))
	string stop   = num2str(ChangePrecision(para[1+mode*2]+para[2+mode*2]/2, 4))
	string units  = WaveUnits(data,1-mode); units[0] = SelectString(strlen(units),""," ")
	string logStr = SelectString(mode,"Horizontal","Vertical")+" profile (Range: "+start+"-"+stop+units+SelectString(scale!=1,"","; scaled by x"+num2str(scale))+") subtracted"
	
	variable yOff = DimOffset(data,1), yDel = DimDelta(data,1)
	variable size = DimSize(data,1-mode), LorR = 0												// default is no shift
	LorR = GetKeyState(0) == 1 ?  1 : LorR														// ctrl/command for positive shift
	LorR = GetKeyState(0) == 2 ? -1 : LorR														// alt/option for negative shift
	
	// +++++++++++++++++++++++++++ find out which wave to use ++++++++++++++++++++++++++++++++++++
	string menuStr = "Current Profile;"+SelectString(WaveExists(cmpL),"","Saved Snapshot;") 
	string loadExt = SelectString(mode,"Horiz","Vert")											// new in ver. 4.45
	SVAR/Z scrPath = stg:$("Load"+loadExt+"WaveLoc")
	wave/Z loadPr  = stg:$("LoadProfile_"+loadExt)
	if (SVAR_Exists(scrPath) && WaveExists(loadPr))
		WaveStats/Q loadPr
		menuStr += SelectString((V_npnts && strlen(scrPath)),"","Loaded Wave;")
	endif
	if (ItemsInList(menuStr)>1)
		PopupContextualMenu "(Select One:;"+menuStr
		if (!V_Flag)
			return ""
		endif
		StrSwitch (S_selection)
			case "Saved Snapshot":
				wave secL = stg:CompareLine
				logStr = SelectString(mode,"Horizontal","Vertical")+" snapshot"+SelectString(scale!=1,""," (scaled by x"+num2str(scale)+")")+" subtracted"
			break
			case "Loaded Wave":
				wave secL = stg:$("LoadProfile_"+loadExt)
				logStr = SelectString(mode,"Horizontal","Vertical")+" external wave (" +ParseFilePath(0,scrPath,":",1,0)+SelectString(scale!=1,"","; scaled by x"+num2str(scale))+ ") subtracted"
			break
		EndSwitch
	endif
	// ++++++++++++++++++++++++++++++++++ do the subtraction +++++++++++++++++++++++++++++++++++++
	if (scale!=1)
		FastOP secL = (scale)*secL
	endif
	
	if (mode == 0 && LorR != 0)																	// subtracts along the Y axis (increasing/decreasing with Y values)
		variable xOff = DimOffset(data,0), xDel = DimDelta(data,0), xSize = DimSize(data,0)
		variable add = ceil(size*abs(yDel/xDel) )												// Add points at both sides to stay within bounds
		Make/FREE/D/N=(xSize+2*add,1) temp = NaN
		MultiThread temp[add,add+xSize-1] = secL[p-add]											// fill with extended profile (endpoints will be repeated)
		SetScale/P x, xOff-xDel*add - LorR*(para[1]-yOff), xDel , temp							// new offset: add extended range and offset from current Y-center (para[1])
		MatrixFilter NanZapMedian temp
	endif
	
	for (i = 0; i < size; i++)
		if (mode)
			MultiThread data[i][][layer] -= secL[q]
		else
			if(WaveExists(temp))
				MultiThread data[][i][layer] -= temp(x-LorR*yDel*i)								// subtract a moving profile
			else
				MultiThread data[][i][layer] -= secL[p]											// subtract stationary profile
			endif
		endif
	endfor
	
	return logStr + SelectString(WaveDims(data)>2,""," from layer "+num2str(layer)) + "."
end

//################################################################################################

static function/S Profile_DivideByProfile(string pName)
	DFREF stg = getWorkFolder(pName)
	SVAR   wLoc = stg:InwaveLocation
	wave   secL = stg:SectionLine
	wave/Z cmpL	= stg:CompareLine
	wave   para = stg:CurrentLines
	wave   data = $wLoc
	
	ControlInfo/W=$pName SetLayer;	 		 int layer = V_flag==5 ? V_Value : 0
	ControlInfo/W=$pName NormSmooth; variable smthFact = V_Flag==5 ? V_Value : kNormSmooth
	
	variable ctrlKey = ((GetkeyState(0) & 2^0) != 0)
	int i, mode = para[0]
	string start  = num2str(ChangePrecision(para[1+mode*2]-para[2+mode*2]/2, 4))
	string stop   = num2str(ChangePrecision(para[1+mode*2]+para[2+mode*2]/2, 4))
	string units  = WaveUnits(data,1-mode); units[0] = SelectString(strlen(units),""," ")
	string logStr = "Normalized to "+SelectString(ctrlKey,"","(exact) ") + SelectString(mode,"horizontal","vertical")+" profile (Range: "+start+"-"+stop+units+")"
	
	// +++++++++++++++++++++++++++ find out which wave to use ++++++++++++++++++++++++++++++++++++
	string menuStr = "Current Profile;"+SelectString(WaveExists(cmpL),"","Saved Snapshot;") 
	string loadExt = SelectString(mode,"Horiz","Vert")											// new in ver. 4.45
	SVAR/Z scrPath = stg:$("Load"+loadExt+"WaveLoc")
	wave/Z loadPr  = stg:$("LoadProfile_"+loadExt)
	if (SVAR_Exists(scrPath) && WaveExists(loadPr))
		WaveStats/Q loadPr
		menuStr += SelectString((V_npnts && V_sum!=0 && strlen(scrPath)),"","Loaded Wave;")
	endif
	if (ItemsInList(menuStr)>1)
		PopupContextualMenu "(Select One:;"+menuStr
		if (!V_Flag)
			return ""
		endif
		StrSwitch (S_selection)
			case "Saved Snapshot":
				wave secL = stg:CompareLine
				logStr = "Normalized to "+SelectString(mode,"Horizontal","Vertical")+" snapshot"
			break
			case "Loaded Wave":
				wave secL = stg:$("LoadProfile_"+loadExt)
				logStr = "Normalized to "+SelectString(mode,"Horizontal","Vertical")+" external wave (" +ParseFilePath(0,scrPath,":",1,0)+")"
			break
		EndSwitch
	endif
	// +++++++++++++++++++++++++++++++++++ do the division +++++++++++++++++++++++++++++++++++++++
	MatrixOP/Free tempSec=replace(secL,0,inf)													// the profile should not get zero => have a big number instead to suppress this row
	if( !ctrlKey && smthFact > 0)																// ctrl / cmd = no smoothing
		Smooth smthFact, tempSec																// smooth the profile a bit
		logStr += " with smoothing (factor " + num2str(smthFact) + ")"
	endif
	for (i = 0; i < DimSize(data,1-mode); i++)
		if (mode)
			MultiThread data[i][][layer] /= tempSec[q]
		else
			MultiThread data[][i][layer] /= tempSec[p]	
		endif
	endfor
	
	return logStr + SelectString(WaveDims(data)>2,""," from layer "+num2str(layer)) + "."
end

//################################################################################################

static function/S Profile_DivideByValue(string pName, string ctrlName, variable &intSum)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc
	
	ControlInfo/W=$pName SetLayer;	 		 int layer = V_flag==5 ? V_Value : 0
	ControlInfo/W=$pName NormSmooth; variable smthFact = V_Flag==5 ? V_Value : kNormSmooth
	
	variable ctrlKey = ((GetkeyState(0) & 2^0) != 0)
	int i, mode = para[0], doArea = !CmpStr(ctrlName,"bNormArea")
	string logStr = "Normalized to the "+SelectString(doArea,"maximum", "area")+" of each "+SelectString(mode,"vertical column","horizontal row")
	if( !ctrlKey && smthFact > 0)																// ctrl / cmd = no smoothing
		logStr += " with smoothing (factor " + num2str(smthFact) + ")"
	endif
	
	// +++++++++++++++++++++++++++++++++++ do the division +++++++++++++++++++++++++++++++++++++++
	variable val, size = DimSize(data,1-mode)
	Make/Free/N=(size) tempPr
	
	for (i = 0; i < DimSize(data,mode); i++)
		if (mode)
			MultiThread tempPr = data[p][i][layer]
		else
			MultiThread tempPr = data[i][p][layer]
		endif
		
		if( !ctrlKey && smthFact > 0)															// ctrl / cmd = no smoothing
			Smooth smthFact, tempPr																// smooth the profile
		endif
		if (doArea)
			MatrixOP/Free tempPr=replace(tempPr,NaN,0)
			SetScale/P x DimOffset(data,1-mode), DimDelta(data,1-mode), WaveUnits(data,1-mode), tempPr	
			val = area(tempPr)
		else
			WaveStats/Q tempPr; val = V_max
		endif
		val = val==0 || numtype(val) ? 1 : val
		intSum += val
		
		if (mode)
			MultiThread data[][i][layer] /= val
		else
			MultiThread data[i][][layer] /= val
		endif
	endfor
	intSum /= DimSize(data,mode)
	
	return logStr + SelectString(WaveDims(data)>2,""," from layer "+num2str(layer)) + "."
end

//################################################################################################

static function/S Profile_Align(string pName)
	DFREF stg = getWorkFolder(pName)
	DFREF sav = GetDataFolderDFR()
	SVAR wLoc = stg:InwaveLocation
	wave para = stg:CurrentLines
	wave data = $wLoc
	int  mode = para[0]

	wave/Z backup = stg:TempCopy																// create an extra save of the current state
	if (!WaveExists(backup))
		Duplicate data, stg:TempCopy
		wave backup = stg:TempCopy
	endif
	
//	int timerRef = StartMSTimer																	// debug
	variable size = DimSize(data,1-mode)
	variable line = DimSize(data,mode)
	variable rev  = (GetkeyState(0) & 2^2) != 0 ? 1 : 0											// shift = reverse sorting 0 = no, 1 = yes
	ControlInfo/W=$pName SetLayer;			int layer = abs(V_flag) == 5 ? V_Value : 0
	ControlInfo/W=$pName cAlignSmooth;		int doSmth = V_Value == 1
	ControlInfo/W=$pName cAlignOutput;		int output = V_Value == 1
	ControlInfo/W=$pName cAlignEdges;		int edgeAlign = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pName cAlignSubShift; 	int subShift  = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pName cAlignFillGaps; 	int fillGaps  = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pName SmoothFactor;		variable smthFact = abs(V_Flag) == 5 ? V_Value : ceil(line*kAlignSmooth/100)
	
	if (output)																					// writes align shift into a file
		Make/O/D/N=(size) $(NameOfWave(data)+"_shft")/Wave=out
		SetScale/P x DimOffset(data,1-mode), DimDelta(data,1-mode), WaveUnits(data,1-mode), out	
	endif

	variable start = 0, stop = line-1, count													// use the cursors to select a sub-region for alignment
	if (strlen(CsrInfo(A)) > 0)
		start = mode ? qcsr(A,"") : pcsr(A,"")
		start = numtype(start) == 0 ? start : 0
	endif
	if (strlen(CsrInfo(B)) > 0)
		stop = mode ? qcsr(B,"") : pcsr(B,"")
		stop = numtype(stop) == 0 ? stop : line-1
	endif
	if (start > stop)
		count = start
		start = stop
		stop  = count
	endif
	
	string logStr = SelectString(subShift,"Aligned ","Sub-shift aligned ")+SelectString(edgeAlign,"","edges ")+"to the "
	Make/FREE/N=(line,1) ShiftWorkA = 0, ShiftWorkB = 0, tempWave = 0
	if (mode)
		ShiftWorkA[start,stop] = backup[0 + rev*(size-1)][p][layer]
		logStr += SelectString(rev,"left","right") + " side"
	else
		ShiftWorkA[start,stop] = backup[p][0 + rev*(size-1)][layer]
		logStr += SelectString(rev,"bottom","top")
	endif
	
	if(doSmth && smthFact > 0)
		Smooth smthFact, ShiftWorkA
		logStr += " with smoothing (factor " + num2str(smthFact) + ")"
	endif
	if (edgeAlign)																				// align the derivative instead
		Differentiate ShiftWorkA/D=tempWave; ShiftWorkA = tempWave
	endif

	if (subShift)
		DebuggerOptions; int sav_debug = V_debugOnError
		DebuggerOptions debugOnError=0 															// switch this off in case the fit fails
	endif
	
	SetDataFolder stg
	data[][][layer] = NaN
	for (count = 0; count < size; count++)														// find the correlation and shift for each row/column
		if (mode)
			ShiftWorkB = backup[count][p][layer]
		else
			ShiftWorkB = backup[p][count][layer]
		endif
		
		if (subShift)
			Duplicate/Free ShiftWorkB, currLine
		endif
		if(doSmth && smthFact > 0)
			Smooth smthFact, ShiftWorkB
		endif
		if (edgeAlign)																			// align the derivative instead
			Differentiate ShiftWorkB/D=tempWave; ShiftWorkB = tempWave
		endif
		
		if (start > 0)																			// limit range for alignment
			ShiftWorkB[0,start-1] = 0
		endif
		if (stop < line-1)
			ShiftWorkB[stop+1,line-1] = 0
		endif
		
		variable align = SpectrumCorrelationShift(ShiftWorkA, ShiftWorkB, subShift)
		
		tempWave = NaN
		variable shiftP1 = -floor(align)*(align < 0)
		variable shiftP2 =  -ceil(align)*(align >=0)
		if (mode)
			if (subShift)		
				tempWave[shiftP1,line-1+shiftP2] = currLine(x+align)
				if (fillGaps)
					MatrixFilter NanZapMedian tempWave
				endif
				data[count][][layer] = tempWave[q]												// write interpolated result
			else
				data[count][shiftP1,line-1+shiftP2][layer] = backup[count][q+align][layer]
			endif
		else
			if (subShift)
				tempWave[shiftP1,line-1+shiftP2] = currLine(x+align)
				if (fillGaps)
					MatrixFilter NanZapMedian tempWave
				endif
				data[][count][layer] = tempWave[p]
			else
				data[shiftP1,line-1+shiftP2][count][layer] = backup[p+align][count][layer]
			endif
		endif
		
		if(output)
			out[count] = DimDelta(data, mode) * align
		endif
	endfor
	SetDataFolder sav
	
	if (subShift)
		DebuggerOptions debugOnError=sav_debug
	endif
	
//	Print "Align shift calculation took", StopMSTimer(timerRef)/1000, "microseconds"			// debug
	if (output)
		logStr += " => output to " + NameOfWave(out)
	endif
	return logStr+"."
end

//+++++++++++++++++++++++++++++++++++++ correlation function +++++++++++++++++++++++++++++++++++++

static function SpectrumCorrelationShift(wave in1, wave in2, variable sub)						// Helper function: Finds maximum correlation of two slices
	Duplicate/Free in2, work
	Correlate/AUTO in1, work
	WaveStats/Q work
	variable error = 1, result = 0																// result = 0 means perfect match
	variable optLoc = sub ? V_maxLoc : (V_maxRowLoc - x2pnt(work, 0))							// if sub-shift is active, return in scaled wave units
	if (V_sum != 0)
		if (V_maxRowLoc >= 3 && sub)
			try
				CurveFit/M=0/Q gauss, work[V_maxRowLoc-3,V_maxRowLoc+3]; AbortOnRTE
				wave W_coef
				if (numtype(W_coef[2]) == 0 && W_coef[2] > optLoc-3 && W_Coef[2] < optLoc+3)
					result = W_coef[2]
					error = 0
				endif
			catch
				if (V_AbortCode == -4)
					variable cfError = GetRTError(1)											// 1 to clear the error
//					Print "Error during sub-step alignment:", GetErrMessage(cfError,3)			// debug
				endif
			endtry
		endif
		if (error)
			result = optLoc
		endif
	endif
	return result
end

//################################################################################################

static function Profile_UndoLastUpdate(string pName, string mode, string btn)
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	wave startcopy	= stg:StartCopy
	wave/Z lastcopy = stg:LastCopy
	wave/Z tempcopy	= stg:TempCopy
	wave data = $wLoc

	variable lastLayers = DimSize(data,2)
	StrSwitch(mode)
		case "save":
			Duplicate/O data, stg:LastCopy
			SetWindow $pName userdata(lastAction)=btn											// save last button press
		break
		case "last":
			if (WaveExists(tempcopy))
				Duplicate/O tempcopy, data, stg:LastCopy										// updating lastcopy will prevent loading an older version upon second press
			elseif (WaveExists(lastcopy))
				Duplicate/O lastcopy, data
			else
				Duplicate/O startcopy, data
			endif
			
			string last = GetUserData(pName,"bUndoLast","lastShift")							// checks if shift button should be restored
			if (strlen(last) > 0)
				Profile_ShiftButtonStatus(pName, -str2num(last))
			endif
			Profile_UpdateAxes(pName, -1)														// reset axis orientation
			SetWindow $pName userdata(lastAction)=btn
		break
		case "restore":
			Duplicate/O startcopy, data															// only copy the data back
			KillWaves/Z lastcopy
		break
		case "all":
			Duplicate/O startcopy, data
			KillWaves/Z lastcopy
			Profile_ShiftButtonStatus(pName, 0)
			Profile_UpdateAxes(pName, -2)														// reset axis orientation
		break
	EndSwitch
	KillWaves/Z tempcopy
	
	Button bUndoLast ,win=$pName ,userdata(lastShift)=""										// delete shift-button info
	ControlInfo/W=$pName bNormUser																// backwards compatibility
	if (V_flag == 1)
		Button bNormUser ,win=$pName ,disable=2*(Profile_ActivateUserInfoButton(data)!=1)		// re-enable the button if data is available
	endif
	if (lastLayers != DimSize(data,2))															// undo has changed the number of layers
		Profile_UpdateLayerControls(pName)
	endif
	return 0
end

//################################################################################################

static function Profile_UpdateLayerControls(string pName)										// toggles layer-related controls
	DFREF stg = getWorkFolder(pName)
	SVAR wLoc = stg:InwaveLocation
	ControlInfo/W=$pName SetLayer
	if (abs(V_flag) != 5)																		// the following controls were all introduced later
		return 0
	endif
	variable layer = V_Value, lSize = DimSize($wLoc,2), lLimit = limit(lsize-1,0,inf)
	ModifyImage/W=$getGraph(pName) ''#0 ,plane = limit(layer,0,lLimit)							// select the next best layer
	SetVariable SetLayer ,win=$pName ,disable = 2*(lsize<2) ,value=_NUM:limit(layer,0,lLimit) ,limits={0,lLimit,1}
	Button bSumLayers    ,win=$pName ,disable = 2*(lsize<2)
	ModifyControlList ControlNameList("",";","bLayer_RotX*")+ControlNameList("",";","bLayer_RotY*") ,win=$pName ,disable = 2*(lsize<2)
	
	ControlInfo/W=$pName bDeleteLayer															// backwards compatibility - added delete button in update
	if (V_flag == 1)
		Button bDeleteLayer   ,win=$pName ,disable = 2*(lsize<2)
	endif
	ControlInfo/W=$pName MultiProfile															// backwards compatibility - added rotate controls and multi-layer in update	
	if (V_flag == 2)
		CheckBox MultiProfile ,win=$pName ,disable = 2*(lsize<2 || lsize>kMaxMultiProfiles)
	endif
	ControlInfo/W=$pName doZProfiles
	if (V_flag == 2)
		int isDisabled = (lsize<2), wasDisabled = (V_disable==2)
		CheckBox doZProfiles  ,win=$pName ,disable = 2*isDisabled
		if ((wasDisabled && !isDisabled) || (!wasDisabled  && isDisabled))						// change of state
			Profile_ExecuteCheckboxAction(pName, "doZProfiles", (V_Value && !isDisabled))
		endif
	endif

	wave/Z multi = stg:MultiSection
	if (WaveExists(multi))																		// make sure the multi-profile has the correct size immediately
		Redimension/D/N=(-1, lSize) multi
	endif
	Profile_AppendSectionTrace(pName)															// do multi-trace updates
	return 0
end